/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include <string.h>

#include "types/void.h"
#include "nxp_kernel.h"
#include "functions.h"
#include "i18n.h"

static int about_found = FALSE;


/* sample[expression, variable, range_start, range_end, samples[, ignore_errors]]
   Returns a vector of SAMPLES samples from EXPRESSION where VARIABLE
   takes values in the range RANGE_START to RANGE_END */
NxpElement* stdlib_eval_sample(NxpElement *el,
			       NxpElementVariable *var,
			       NxpElementReal *r_start,
			       NxpElementReal *r_end,
			       NxpElementInt *n,
                               GSList *rest_of_args,
                               ETC)
{
    NxpElementVector *result;
    NxpElement *expression, *tmp;
    NxpIdentifier *variable;
    gint samples, i;
    double range_start, range_end, t, k;
    gboolean ignore_errors = FALSE;

    expression = el;
    variable = var->ident;
    range_start = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(r_start));
    range_end = nxp_element_r_get_value_as_double(NXP_ELEMENT_R(r_end));

    if (range_start > range_end) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("function 'sample': value error: RANGE START > RANGE END."));
	return NULL;
    }

    samples = nxp_element_int_get_si(NXP_ELEMENT(n), error);

    if (samples < 0) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("argument 5 of function 'sample' (SAMPLES) should not be negative."));
	return NULL;
    }

    switch (g_slist_length(rest_of_args))
    {
    case 0:
        break;
    case 1: {
	GError *error1 = NULL;
        tmp = NXP_ELEMENT(rest_of_args->data);
        ignore_errors = nxp_element_is_true(tmp, context, &error1);
        if (error1) {
	    g_propagate_error(error, error1);
            return NULL;
        }
        break;
    }
    default:
        g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN,
		    _("Too many arguments to 'sample'."));
        return NULL;
    }
    result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(samples, NULL));
    if (samples == 0)
	return NXP_ELEMENT(result);
    /* create anonymous name space so that variables defined here will
       not interfere with existing ones. */
    nxp_eval_context_push_name_space(context);
    k = (range_end - range_start) / (samples - 1);
    for (i = 0; i < samples; i++)
    {
	NxpElement *sample;
	GError *error1 = NULL;
	t = range_start + k * i;
	tmp = nxp_element_real_new(t);
	nxp_user_variable_define(variable, tmp, context, &error1);
	NXP_ELEMENT_UNREF(tmp);
	if (error1) {
	    NXP_ELEMENT_UNREF(result);
	    nxp_eval_context_pop_name_space(context);
	    g_propagate_error(error, error1);
	    return NULL;
	}
	sample = nxp_element_eval(expression, context, &error1);
	if (error1) {
            if (ignore_errors) {
                _element_vector_index(result, i) = nxp_element_void_new();
                g_clear_error(&error1);
                continue;
            }
	    NXP_ELEMENT_XUNREF(sample);
	    NXP_ELEMENT_UNREF(result);
	    nxp_eval_context_pop_name_space(context);
	    g_propagate_error(error, error1);
	    return NULL;
	}
	_element_vector_index(result, i) =  sample;
    }
    nxp_eval_context_pop_name_space(context);
    return NXP_ELEMENT(result);
}





static void about_ghfunc(gpointer key, gpointer value, const gpointer user_data)
{
    struct {
	NxpContext *context; const char *name; guint size;
    } *data = user_data;
    GSList* l;
    GString* tmp;

    if (strncmp((char*)key, data->name, data->size) == 0) {
	NxpDefUserFunc* def = (NxpDefUserFunc*)value;

	about_found = TRUE;

	nxp_output_message(data->context, _("about [ %s ]:"), def->definition.name);
	tmp = nxp_element_str(def->expression);
	nxp_output_message(data->context, _("\texpression: %s"), tmp->str);
	nxp_output_message(data->context, _("\targument list:"));
	for (l=def->argument_list; l; l=g_slist_next(l)) {
	    nxp_output_message(data->context, "\t\t%s", (char*)l->data);
	}
	g_string_free(tmp, TRUE);
    }
}

/* Simple, but does the work! */
/* Later, we should change &lt; and &gt; to < and >... */
static char* removeTags(char* str) {
    char *start_point = str;
    char *read_point  = str;
    char *write_point = str;

    gboolean inTag = 0;
    while (*read_point) {
	if (inTag) {
	    if (*read_point == '>') 
		inTag = 0;
	} else
	    if (*read_point == '<')
		inTag = 1;
	    else {
		*write_point = *read_point;
		write_point++;
	    }
	read_point++;
    }
    *write_point = '\0';
    return start_point;
}

static NxpElement* topic_help(NxpElementString *topic, ETC)
{
    char *help = NULL;
    help = nxp_help(topic->string, context);
    help = removeTags(help);
    nxp_output_message(context, "%s", help);

    return nxp_element_void_new();

}

static NxpElement* help(NxpElementVariable* id, ETC)
{
    GString *fullpath = NULL;
    const char *name = NULL;
    NxpNameSpace *ns = NULL;
    NxpDefinition *def = NULL;

    nxp_identifier_resolve(id->ident, context, &ns, &name);
    if (ns) {
	fullpath = nxp_name_space_get_full_path_name(ns);
	if (name == NULL) {
	    char *help = NULL;
	    help = nxp_help(nxp_identifier_to_string(id->ident),
                            context);
	    help = removeTags(help);
	    nxp_output_message(context, "%s", help);
	} else {
	    def = g_hash_table_lookup(ns->definitions, name);
	    if (def) {
		char *help = NULL;
		fullpath = g_string_append(fullpath, name);
		help = nxp_help(fullpath->str, context);
		g_string_free(fullpath, TRUE);
		help = removeTags(help);
		nxp_output_message(context, "%s", help);
	    } else {
		nxp_output_message(context, _("No help about '%s'"), nxp_identifier_to_string(id->ident));
	    }
	}
    } else {
	nxp_output_message(context, _("No help about '%s'"), nxp_identifier_to_string(id->ident));
    }
    
    return nxp_element_void_new();
}

static NxpElement* about(NxpElementVariable* id, ETC)
{
    const char* name = NULL;
    char *tmp_name;
    NxpNameSpace* ns = NULL;
    NxpIdentifier *ident = id->ident;
    NxpDefinition *def;
    GString *tmp;

    /* Resolve identifier */
    nxp_identifier_resolve(ident, context, &ns, &name);

    /* it's a namespace? */
    if (name == NULL) {
	nxp_output_message(context, _("That's a namespace!"));
	return nxp_element_void_new();
    }
    
    /* it exists? */

    def = g_hash_table_lookup(ns->definitions, name);
    if (!def) {
	tmp_name = g_strconcat(name, "/", NULL);
	struct {
	    NxpContext *context; const char *name; guint size;
	} data = { context, tmp_name, strlen(tmp_name) };
	/* We must check if it is a user defined function */
	g_hash_table_foreach(ns->definitions, about_ghfunc, &data);

	g_free(tmp_name);
	/* if it is not... */
	if (about_found) {
	    about_found = FALSE;
	} else {
	    nxp_output_message(context, _("I don't know what's '%s'!"), name);
	}
	return nxp_element_void_new();
    }

    /* yes... let's see what it is! */
    nxp_output_message(context, _("about [ %s ]:"), name);
    switch(def->type) {
    case NXP_DEF_VARIABLE:
	tmp = nxp_element_str(((NxpDefVariable*)def)->value);
	nxp_output_message(context, _("\ttype: variable"));
	nxp_output_message(context, _("\tvalue: %s"), tmp->str);
	g_string_free(tmp, TRUE);
	break;
    case NXP_DEF_BI_FUNC:
	nxp_output_message(context, _("\ttype: built-in function"));
	{
	    NxpDefBIFunc* bi = (NxpDefBIFunc*)def;
	    GSList* l;
	    for( l = bi->prototypes; l ;l = g_slist_next(l)) {
		int n;
		NxpProtoInstance* npi = (NxpProtoInstance*)l->data;
		if (npi->eval_func) {
		    const NxpProtoClass* pc = npi->protoclass;
		    nxp_output_message(context, _("\tarity: %d"), pc->arity);
		    for (n=0; n<pc->arity; n++) {
			if (pc->types[n] == NXP_ELEMENT_TYPE_ANY) {
			    nxp_output_message(context, "\t\t(%d) %s",
					       pc->types[n],_("any expression"));
			} else 
			    nxp_output_message(context, "\t\t(%d) %s", 
					       pc->types[n],
					       nxp_element_type_get_info(pc->types[n])->name);
		    }
		}
	    }
	}
	break;
    case NXP_DEF_USER_FUNC:
	/* Catched earlier... */
	break;
    default:
	nxp_output_message(context, _("\ttype: unkown!"));
    }

    return nxp_element_void_new();
}


NxpElement* bifunc_eval_version(ETC)
{
    nxp_output_message(context, _("NumExp version %s"), nxp_kernel_version());
    nxp_output_message(context, _("(C) 1998-2005 The NumExp Team"));
    return nxp_element_void_new();
}


NxpElement* bifunc_eval_debug_info(ETC)
{
    GString *ns_path;
    ns_path = nxp_name_space_get_full_path_name(context->cwd);
    nxp_output_message(context, _("Debugging info:\n" "context->cwd = %s\n"), ns_path->str);
    g_string_free(ns_path, TRUE);
    return nxp_element_void_new();
}


static void bifunc_eval_ls_ghfunc(gpointer key, gpointer value, gpointer user_data)
{
    NxpDefinition *def;
    struct {
	NxpContext *context;
	int         flag;
    } *data = user_data;

    g_return_if_fail(key && value);
    def = value;
    if (def->type == data->flag) {
	nxp_output_message(data->context, "\t- %s", (char*)key);
    }
}


NxpElement* bifunc_eval_ls(GSList *args, ETC)
{
    GString *ns_path;
    NxpNameSpace *ns, *child;
    const char *name;
    NxpElement *arg;
    struct {
	NxpContext *context;
	int         flag;
    } data = { context, 0 };

    if (args) {
	arg = args->data;
	g_return_val_if_fail(arg, NULL);
	if (NXP_ELEMENT_GET_TYPE(arg) == NXP_ELEMENT_TYPE_VARIABLE) {
	    NxpIdentifier *ident = NXP_ELEMENT_VARIABLE(arg)->ident;
	    nxp_identifier_resolve(ident, context, &ns, &name);
	} else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, _("ls: type error -- expected path"));
	    return NULL;
	}
    } else {
	ns = context->cwd;
    }
    g_return_val_if_fail(ns, NULL);
    nxp_name_space_lock(ns);
    ns_path = nxp_name_space_get_full_path_name(ns);

    nxp_output_message(context, _("Listing of name space [ %s ]:"), ns_path->str);
    g_string_free(ns_path, TRUE);    

    nxp_output_message(context, _("Variables"));
    data.flag = NXP_DEF_VARIABLE;
    g_hash_table_foreach(ns->definitions, 
			 bifunc_eval_ls_ghfunc, &data);

    nxp_output_message(context, _("\nNumExp functions:"));
    data.flag = NXP_DEF_BI_FUNC;
    g_hash_table_foreach(ns->definitions, 
			 bifunc_eval_ls_ghfunc, &data);

    nxp_output_message(context, _("\nUser defined functions:"));
    data.flag = NXP_DEF_USER_FUNC;
    g_hash_table_foreach(ns->definitions, 
			 bifunc_eval_ls_ghfunc, &data);

    nxp_output_message(context, _("\nChild namespaces:"));

    child = ns->child;
    for (; child; child = child->next) {
	nxp_name_space_lock(child);
	//ns_path = nxp_name_space_get_full_path_name(child);
	nxp_output_message(context, "\t%s", child->name);
	//g_string_free(ns_path, TRUE);
	nxp_name_space_unlock(child);
    }

    nxp_name_space_unlock(ns);
    return nxp_element_void_new();
}

// This is defined somewhere else...
// -- numexpc implements its own;
// -- bonobo server implements the corba one;
void Numexp_Kernel_to_Client_cwd_change(NxpEvalContext *context);

NxpElement* bifunc_eval_cd(NxpElementVariable* str, ETC)
{
    NxpIdentifier *ident;
    NxpNameSpace *ns, *child;
    const char* name = NULL;

    ident = NXP_ELEMENT_VARIABLE_IDENTIFIER(str);
    ns = context->cwd;
    nxp_identifier_resolve(ident, context, &child, &name);

    /* The four next lines are useful because it is long to type a directory
     * name followed by the path separator */
    if (child && name) {
	nxp_identifier_append_separator(ident);
	nxp_identifier_resolve(ident, context, &child, &name);
    }

    if (child && !name)
	context->cwd = child;
    else {
	gchar *id = nxp_identifier_to_string(ident);
	if (!name)
	    g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED,
			_("Namespace '%s' does not exist"), id);
	g_free(id);
	return NULL;
    }
    /* Send message to the childrens */
    Numexp_Kernel_to_Client_cwd_change(context);
    return nxp_element_void_new();
}

NxpElement* bifunc_eval_pwd(ETC)
{
    GString *str;
    NxpElement *el;

    str = nxp_name_space_get_full_path_name(context->cwd);
    el = nxp_element_string_new(str->str);
    g_string_free(str, TRUE);
    return el;
}

#define D(str, prot, eval, derive, simp) nxp_bifunc_define_global(str, \
                                         prot, eval, derive, simp, context, &err);\
                                         if (err) { g_propagate_error(error, err); return; }

#define DV(str, prot, eval, derive, simp) nxp_bifunc_define_global_volatile(str, \
                                          prot, eval, derive, simp, context, &err);\
                                          if (err) { g_propagate_error(error, err); return; }

static void bifunc_initialize(NxpEvalContext *context, GError **error)
{
    GError *err = NULL;

    nxp_built_in_variable_define_simple("void", nxp_element_void_new(), context, &err);
    if (err) { g_propagate_error(error, err); return; }


    D("replaceOnVector", &nxp_prototype__VARIABLE_INT_ANY, 
      bifunc_eval_replace_on_vecvar, NULL, bifunc_eval_replace_on_vecvar);
    D("replaceOnVector", &nxp_prototype__VECTOR_INT_ANY, 
      bifunc_eval_replace_on_vector, NULL, bifunc_eval_replace_on_vector);

    D("getFromMatrix", &nxp_prototype__MATRIX_INT_INT,
      bifunc_eval_get_from_matrix, NULL, bifunc_eval_get_from_matrix);

    D("replaceOnMatrix", &nxp_prototype__VARIABLE_INT_INT_ANY,
      bifunc_eval_replace_on_matvar, NULL, bifunc_eval_replace_on_matvar);
    D("replaceOnMatrix", &nxp_prototype__MATRIX_INT_INT_ANY,
      bifunc_eval_replace_on_matrix, NULL, bifunc_eval_replace_on_matrix);

    D("define", &nxp_prototype__FUNCTION_ANY, 
      bifunc_eval_define_function, NULL, bifunc_eval_define_function);

    D("undef", &nxp_prototype__VARIABLE,
      bifunc_eval_undef_function, NULL, bifunc_eval_undef_function);

    D("round", &nxp_prototype__R, 
      bifunc_eval_round, NULL, bifunc_eval_round);

    D("round", &nxp_prototype__REAL_INT, 
      bifunc_eval_round2, NULL, bifunc_eval_round2);

    D("mkdir", &nxp_prototype__VARIABLE, 
      bifunc_create_world, NULL, bifunc_create_world);

    D("debugInfo", &nxp_prototype__VOID, 
      bifunc_eval_debug_info, NULL, bifunc_eval_debug_info);

    DV("version", &nxp_prototype__VOID, 
       bifunc_eval_version, NULL, bifunc_eval_version);

    D("quote", &nxp_prototype__ANY, 
      bifunc_eval_quote, NULL, bifunc_eval_quote);

    D("if", &nxp_prototype__ANY_ANY, 
      bifunc_eval_if2, NULL, bifunc_simp_if2);

    D("if", &nxp_prototype__ANY_ANY_ANY, 
      bifunc_eval_if, NULL, bifunc_simp_if); 

    D("D", &nxp_prototype__ANY_VARIABLE, 
      bifunc_eval_d, NULL, bifunc_simplify_d);

    D("D", &nxp_prototype__ANY_VARIABLE_INT, 
      bifunc_eval_n_d, NULL, bifunc_simplify_n_d);

    D("load", &nxp_prototype__STRING, 
      bifunc_load, NULL, bifunc_load);

    D("saveSession", &nxp_prototype__STRING,
      bifunc_save_session, NULL, bifunc_save_session);

    D("loadSession", &nxp_prototype__STRING,
      bifunc_load_session, NULL, bifunc_load_session);

    D("ls", &nxp_prototype__ELLIPSES, 
      bifunc_eval_ls, NULL, bifunc_eval_ls);

    D("cd", &nxp_prototype__VARIABLE,
      bifunc_eval_cd, NULL, bifunc_eval_cd);

    D("pwd", &nxp_prototype__VOID,
      bifunc_eval_pwd, NULL, bifunc_eval_pwd);

    D("print", &nxp_prototype__ELLIPSES,
      bifunc_eval_print, NULL, bifunc_simp_print);

    D("prog2", &nxp_prototype__ELLIPSES,
      bifunc_eval_prog2, NULL, bifunc_eval_prog2);

    D("prog1", &nxp_prototype__ELLIPSES,
      bifunc_eval_prog1, NULL, bifunc_eval_prog1);

    D(":", &nxp_prototype__ELLIPSES,
      bifunc_eval_two_dot, NULL, bifunc_simplify_two_dot);

    D("progn", &nxp_prototype__ELLIPSES,
      bifunc_eval_two_dot, NULL, bifunc_simplify_two_dot);

    D("while", &nxp_prototype__ELLIPSES,
      bifunc_eval_while, NULL, bifunc_simp_while);   

    D("for", &nxp_prototype__VARIABLE_VECTOR_ANY,
      bifunc_eval_for, NULL, bifunc_simp_for);

    D("factorial", &nxp_prototype__INT,
      bifunc_eval_factorial, NULL, bifunc_eval_factorial);

    D("isReal", &nxp_prototype__ANY,
      bifunc_eval_is_real, NULL, NULL);

    D("isInt", &nxp_prototype__ANY,
      bifunc_eval_is_int, NULL, NULL);

    D("isComplex", &nxp_prototype__ANY,
      bifunc_eval_is_complex, NULL, NULL);

    D("isVector", &nxp_prototype__ANY,
      bifunc_eval_is_vector, NULL, NULL);

    D("isMatrix", &nxp_prototype__ANY,
      bifunc_eval_is_matrix, NULL, NULL);

    D("simplify", &nxp_prototype__ANY,
      bifunc_eval_simplify, NULL, NULL);

    D("eval", &nxp_prototype__ANY,
      bifunc_eval_eval, NULL, bifunc_eval_eval);

    D("about", &nxp_prototype__VARIABLE,
      about, NULL, about);

    D("help", &nxp_prototype__VARIABLE,
      help, NULL, help);

    D("help", &nxp_prototype__STRING,
      topic_help, NULL, topic_help);

    D("map", &nxp_prototype__ANY_MATRIX,
      bifunc_eval_map_single_matrix, NULL, bifunc_simp_map_single_matrix);

    D("map", &nxp_prototype__ANY_VARIABLE_MATRIX,
      bifunc_eval_map_matrix, NULL, bifunc_simp_map_matrix);

    D("map", &nxp_prototype__ANY_VECTOR,
      bifunc_eval_map_single_vector, NULL, bifunc_simp_map_single_vector);

    D("map", &nxp_prototype__ANY_VARIABLE_VECTOR,
      bifunc_eval_map_vector, NULL, bifunc_simp_map_vector);

    D("fill", &nxp_prototype__INT_ANY,
      bifunc_eval_vector_fill, NULL, bifunc_simp_vector_fill);

    D("size", &nxp_prototype__VECTOR,
      bifunc_size_vector, NULL, bifunc_size_vector);

    D("size", &nxp_prototype__MATRIX,
      bifunc_size_matrix, NULL, bifunc_size_matrix);

    D("join", &nxp_prototype__VECTOR_VECTOR,
      bifunc_eval_join_vector, NULL, bifunc_simp_join_vector);

    D("join", &nxp_prototype__MATRIX_MATRIX,
      bifunc_eval_join_matrix, NULL, bifunc_simp_join_matrix);

    D("vectorToPoly", &nxp_prototype__VECTOR_VARIABLE,
      bifunc_vector_to_poly, NULL, bifunc_vector_to_poly);

#ifdef HAVE_COMPLEX_H
#define DEF(f,op, derive)       nxp_bifunc_define_global(f, &nxp_prototype__R, \
                                 stdlib_eval_r_##op, derive, NULL, context, NULL); \
                        nxp_bifunc_define_global(f, &nxp_prototype__COMPLEX, \
                             stdlib_eval_complex_##op, derive, NULL, context, NULL);
#else
#define DEF(f,op)   nxp_bifunc_define_global(f, &nxp_prototype__R, \
                             stdlib_eval_r_##op, derive, NULL, context, NULL);
#endif

DEF("ln", ln, stdlib_derive_ln);
nxp_bifunc_define_global("ln", &nxp_prototype__ANY, NULL, stdlib_derive_ln, NULL, context, NULL); 

#undef DEF

#undef D
#undef DV


#define D(str, prot, eval, derive, simp) nxp_bifunc_define_global(str, \
                                         prot, eval, derive, simp, context, &err);\
                                         if (err) { g_propagate_error(error, err); return; }

    D("sample", &nxp_prototype__ANY_VARIABLE_REAL_REAL_INT_ELLIPSES, stdlib_eval_sample, NULL, NULL);
#undef D

}



void nxp_module_init(NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    bifunc_initialize(context, &err);
    if (err) { g_propagate_error(error, err); return; }
}
