/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>

#include "element.h"
#include "message.h"
#include "config.h"
#include "element_type.h"
#include "element_output.h"
#include "eval_context.h"
#include "identifier.h"
#include "bifuncs.h"
#include "types/types.h"
#include "errors.h"
#include "nxp_debug.h"
#include "i18n.h"

NxpElementPreferences	nxp_element_preferences;
gint			nxp_element_trace_references	= 0;
volatile gint		nxp_element_global_count	= 0;
volatile gint		nxp_element_references_count	= 0;

#ifdef NXP_ELEMENT_DEBUG
/* Singly-linked list that keeps track of all allocated elements. */
static GSList		*elements_list			= NULL;
#endif


/* --- this structure contains variables to pass cancelation requests
   to threads (one message at a time only) --- */
static struct {
    pthread_mutex_t mut;
    pthread_cond_t cond;
    pthread_t tid;
} thread_cancelation_request = 
{
    PTHREAD_MUTEX_INITIALIZER,
    PTHREAD_COND_INITIALIZER,
    0
};

/* ----------------- END OF VARIABLES AND STRUCTURES ----------------------- */


/**
 * nxp_element_abort_thread:
 * @tid: thread id (on linux it's also a pid) of the thread you wish to abort.
 * 
 * Send a message to a thread running nxp_element_eval() requesting it to
 * abort calculations and quit as soon as possible, returning
 * NULL. Only the main thread should call this function.
 **/
void nxp_element_abort_thread(pthread_t tid)
{
    struct timeval now;
    struct timespec timeout;
    int retcode;

    g_assert(thread_cancelation_request.tid == 0);
    
    pthread_mutex_lock(&thread_cancelation_request.mut);

    gettimeofday(&now, NULL);
    timeout.tv_sec = now.tv_sec + 5; /* wait 5 seconds for thread to finish */
    timeout.tv_nsec = now.tv_usec * 1000;
    retcode = 0;
    thread_cancelation_request.tid = tid;
    while (thread_cancelation_request.tid != 0 && retcode != ETIMEDOUT)
    {
	retcode = pthread_cond_timedwait(&thread_cancelation_request.cond, 
					 &thread_cancelation_request.mut,
					 &timeout);
    }
    if (retcode == ETIMEDOUT) {
	g_warning("Eval thread didn't respond to the exit request!\n"
		  "This is probably a bug, so please send a bug report"
		  " to one of the developers. Thanks!");
    } else {
	g_message("Eval thread aborted by user request.");
    }
    pthread_mutex_unlock(&thread_cancelation_request.mut);
}




gboolean  nxp_element_check_thread_abortion_request(void)
{
    gboolean rv;
    pthread_mutex_lock(&thread_cancelation_request.mut);
    if (pthread_equal(thread_cancelation_request.tid, pthread_self()))
    {
	thread_cancelation_request.tid = 0;
	pthread_cond_broadcast(&thread_cancelation_request.cond);
	rv = TRUE;
    } else
	rv = FALSE;
    pthread_mutex_unlock(&thread_cancelation_request.mut);
    pthread_testcancel();
    return rv;
}

gboolean nxp_element_check_type(NxpElement *el, int type)
{
    NxpElementTypeInfo *typeinfo = NULL;

    if (el) {
	typeinfo = el->type_info;
	while (typeinfo && typeinfo->type != type)
	    typeinfo = typeinfo->parent;
    }
    return (typeinfo != NULL);
}




#ifdef NXP_ELEMENT_DEBUG

# define NXP_ELEMENT_MAGIC ((guint16)12345)
# define NXP_ELEMENT_DELETED_MAGIC ((guint16)54321)


NxpElement *nxp__element_typecast_and_sanity_check(gpointer p)
{
    NxpElement *el;
    if (p == NULL) {
	fprintf(stderr, "WARNING: Cast from NULL pointer to (NxpElement *)\n");
	if(_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	    G_BREAKPOINT();
	return (NxpElement *)p;
    }
    if (GPOINTER_TO_UINT(p) < 0x100) {
	fprintf(stderr, "WARNING: address %p looks suspiciously low!\n", p);
	if(_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	    G_BREAKPOINT();
    }
    el = (NxpElement *)p;
    if (el->magic != NXP_ELEMENT_MAGIC) {
	fprintf(stderr, "WARNING: Cast from invalid pointer to (NxpElement*)"
		" (pointer = %p, magic = %u)\n", el, el->magic);
	if(_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	    G_BREAKPOINT();
    }
    return el;
}


void nxp__element_ref(NxpElement *el, const char *file, const char *function, int line)
{
    NxpElementRefTraceItem *trace;
    if (!el) {
	fprintf(stderr, "Warning: REFerencing NULL element\n");
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) G_BREAKPOINT();
	return;
    }
    g_atomic_int_inc(&el->ref_count);
    g_atomic_int_inc(&nxp_element_references_count);
    if (nxp_element_trace_references)
    {
	trace = g_new0(NxpElementRefTraceItem, 1);
	trace->ref = 1;
	trace->file = file;
	trace->function = function;
	trace->line = line;
	el->ref_trace = g_slist_prepend(el->ref_trace, trace);
    }
}


void nxp__element_unref(NxpElement *el, const char *file, const char *function, int line)
{
    NxpElementRefTraceItem *trace;
    if (!el) {
	fprintf(stderr, "Warning: UNREFerencing NULL element\n");
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) G_BREAKPOINT();
	return;
    }
    if (g_atomic_int_dec_and_test(&el->ref_count))
	nxp__element_destroy(el);
    else {
	if (nxp_element_trace_references)
	{
	    trace = g_new0(NxpElementRefTraceItem, 1);
	    trace->ref = 0;
	    trace->file = file;
	    trace->function = function;
	    trace->line = line;
	    el->ref_trace = g_slist_prepend(el->ref_trace, trace);
	}
    }
    --nxp_element_references_count;
}

NxpElement* nxp__element_check_type(NxpElement *el, int type)
{
    const char *name1;
    const char *name2;

    if (!nxp_element_check_type(el, type)) {
	name1 = NXP_ELEMENT_GET_TYPE_NAME(el);
	name2 = nxp_element_type_get_info(type)->name;
	fprintf(stderr, "WARNING: invalid element cast from %s to %s\n", name1, name2);
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	    G_BREAKPOINT();
    }
    return el;
}

#endif /* #ifdef NXP_ELEMENT_DEBUG */





/**
 * nxp_element_alloc:
 * @type_info: pointer to the #NxpElementTypeInfo structure for the
 * type of element you wish to allocate
 * @size: the size of the element structure
 *
 * Allocate memory for an element of type @type_info->type and
 * initialize its common fields. This function is low-level and
 * usually should not be used.
 * 
 * Return value: pointer to the new element.
 **/
NxpElement* nxp_element_alloc(NxpElementTypeInfo *type_info, size_t size)
{
    NxpElement *elem;

    g_assert(type_info != NULL);
    g_assert(size > 0);
    elem = g_malloc0(size);
    elem->type_info = type_info;
    elem->ref_count = 1;

#ifdef NXP_ELEMENT_DEBUG
    elem->magic = NXP_ELEMENT_MAGIC;
    ++nxp_element_global_count;
    ++nxp_element_references_count;
    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	elements_list = g_slist_prepend(elements_list, elem);
#endif

    return elem;
}



void nxp__element_destroy(NxpElement* el)
{
    NxpElementType type;
    NxpElementTypeInfo *typeinfo;
    
    type = NXP_ELEMENT_GET_TYPE(el);
    if (type == NXP_ELEMENT_TYPE_VOID) {
	printf("Attemped to destroy the 'void' element!\n");
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	    G_BREAKPOINT();
	} 
	el->ref_count = 1;
	return;
    }
    if (type < 1) {
	printf("Destroying a not recognized element!\n");
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	    G_BREAKPOINT();
	} 
	return;
    }

    typeinfo = nxp_element_type_get_info(type);
    g_assert(typeinfo);
    if (typeinfo->destroy_func) typeinfo->destroy_func(el);

#ifdef NXP_ELEMENT_DEBUG
    el->magic = NXP_ELEMENT_DELETED_MAGIC;
    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	elements_list = g_slist_remove(elements_list, el);
    if (nxp_element_trace_references) {
	g_slist_foreach(el->ref_trace, (GFunc) g_free, NULL);
	g_slist_free(el->ref_trace);
    }
#endif
    /* el->type = 0; */
    g_free(el);
    --nxp_element_global_count;
}


/**
 * nxp_element_foreach:
 * @elem: the element where to start iterating
 * @func: callback function
 * @data: user data to be added to the function
 * @context: eval context
 * 
 * Iterate over each element in an element tree and call func() on it.
 **/
void nxp_element_foreach(NxpElement *elem,
                         NxpElementVisitorFunc func,
                         gpointer user_data,
                         NxpEvalContext *context)
{
    g_return_if_fail(elem);
    g_return_if_fail(func);

    func(elem, user_data);
    if (elem->type_info->iterator_func)
 	elem->type_info->iterator_func(elem, func, user_data, context);
}


/**
 * nxp_element_depends_of:
 * @self: expression to test
 * @variable: variable
 * @error: error
 *
 * Tests if @self depends or not of @variable.
 * 
 * Return value: a boolean
 **/
gboolean
nxp_element_depends_on(NxpElement *self, NxpElement *variable,
		       NxpEvalContext *context, GError **error)
{
    NxpElementTypeInfo *typeinfo;
    g_return_val_if_fail(self, FALSE);

    typeinfo = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(self));
    
    if (typeinfo->dependson_func)
	return typeinfo->dependson_func(self, variable, context, error);
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("Don't know how to determinate if an element of type '%s' depends on a variable or not."), NXP_ELEMENT_GET_TYPE_NAME(self));
	return FALSE;
    }
}


/**
 * nxp_element_substitute:
 * @el: expression to manipulate
 * @var: variable to substitute with the new expression @el.
 * @el: the expression that replaces @var.
 *
 * Compute a new element that would be equal to @elem if @var is changed with
 * @el. No simplification is made. And we can, for example replace the variable
 * x by x+1. We don't assume any equality (very useful to evaluate a function).
 * 
 * Return value: a new expression.
 **/
NxpElement* nxp_element_substitute	      (NxpElement	       *elem,
					       NxpElement	       *var,
					       NxpElement	       *el,
					       NxpEvalContext	       *context,
					       GError		       **error)
{
    NxpElementTypeInfo *typeinfo;

    g_return_val_if_fail(elem, NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_VARIABLE(var), NULL);

    typeinfo = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(elem));
    g_return_val_if_fail(typeinfo, NULL);
    if (typeinfo->substitute_func) {
	return typeinfo->substitute_func(elem, var, el, context, error);
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("Don't know how to substitute a variable with an expression in an element of type '%s'"), NXP_ELEMENT_GET_TYPE_NAME(elem));
	return NULL;
    }
}

/**
 * nxp_element_simplify:
 * @el: expression to simplify
 * @error: error
 *
 * Compute a simplified expression equivalent to @el.
 * 
 * Return value: a new simplified expression
 **/
NxpElement* nxp_element_simplify(NxpElement* el, NxpEvalContext *context, GError **error)
{
    NxpElementTypeInfo *typeinfo;

    g_return_val_if_fail(el, NULL);
    if (nxp_element_check_thread_abortion_request()) return NULL;

    typeinfo = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(el));
    g_return_val_if_fail(typeinfo, NULL);
    if (typeinfo->simplify_func) {
	GError *error1 = NULL;
	NxpElement *result;
	result = typeinfo->simplify_func(el, context, &error1);
	if (error1) g_propagate_error(error, error1);
	return result;
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("Don't know how to simplify element of type '%s'"), NXP_ELEMENT_GET_TYPE_NAME(el));
	return NULL;
    }
}



/**
 * nxp_element_eval:
 * @el: element to evaluate
 * @error: a possible run-time error
 *
 * Evaluate element tree.
 * 
 * Return value: the result of evaluating @el
 **/
NxpElement* nxp_element_eval(NxpElement* el, NxpEvalContext *context, GError **error) 
{
    NxpElementTypeInfo *typeinfo;
    GError *error1 = NULL;

    g_return_val_if_fail(el, NULL);
    if (nxp_element_check_thread_abortion_request()) return NULL;

    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(el);
    g_return_val_if_fail(typeinfo, NULL);
    if (typeinfo->eval_func) {
	NxpElement *ret;
	ret = typeinfo->eval_func(el, context, &error1);
	if (error1) {
	    g_propagate_error(error, error1);
	}
	return ret;
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("Don't know how to evaluate element of type '%s'"), NXP_ELEMENT_GET_TYPE_NAME(el));
    }
    return NULL;
}



NxpElement* nxp_element_derive_simple(NxpElement* el, const char *var, NxpEvalContext *context) 
{
    NxpIdentifier *ident = nxp_identifier_new_simple(var);
    NxpElement *result = nxp_element_derive(el, ident, context, NULL);
    nxp_identifier_unref(ident);
    return result;
}


NxpElement* nxp_element_ref(NxpElement *elem)
{
    return NXP_ELEMENT_REF(elem);
}


void nxp_element_data_unref (gpointer data) {
    NXP_ELEMENT_UNREF(NXP_ELEMENT(data));
}

NxpElement* nxp_element_unref(NxpElement *elem)
{
    return NXP_ELEMENT_UNREF(elem);
}


/**
 * nxp_element_derive:
 * @el: 
 * @var: 
 * @error:
 * 
 * Derive (in a mathematical sense, differenciate) expression @el.
 * 
 * Return value: 
 **/
NxpElement* nxp_element_derive(NxpElement* el, NxpIdentifier *var, 
			       NxpEvalContext *context, GError **error)
{
    NxpElementTypeInfo *typeinfo;

    g_return_val_if_fail(el, NULL);
    if (nxp_element_check_thread_abortion_request()) return NULL;

    typeinfo = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(el));
    g_return_val_if_fail(typeinfo, NULL);
    if (typeinfo->derive_func) {
	GError *myerror = NULL;
	NxpElement* ele;
	ele = typeinfo->derive_func(el, var, context, &myerror); 
	if (myerror) { g_propagate_error(error, myerror); return NULL; }
	return ele;
    } else
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    "%s", NXP_ELEMENT_GET_TYPE_NAME(el));
    return NULL;
}


/**
 * nxp_elements_init:
 * @void: 
 * 
 * Initialize the elements subsystem.
 **/
void nxp_elements_init(void)
{
/*      int ret; */

    /* Initialize the element types' system */
    nxp_element_type_init();
    nxp_errors_init();
    /* Create TSD key for element contexts */
/*      ret = pthread_key_create(&nxp_element_context_tsd_key, NULL); */
/*      g_assert(ret == 0); */
    /* Initialize parser mutex. */
/*      exp_parser_init();  -- this should be called by applications */
    /* Create global (root) definitions' name space. */
//    nxp_global_name_space = nxp_name_space_new(NULL, "(global_name_space)");
    /* Create the main thread's NxpElementContext. */
/*      main_element_cntxt_master = nxp_element_cntxt_master_new(nxp_global_name_space, NULL); */
/*      main_element_context = nxp_element_context_new(main_element_cntxt_master); */
/*      nxp_element_cntxt_master_switch_to(main_element_cntxt_master); */
/*      nxp_element_context_switch_to(main_element_context); */

    /* Initialize the operator bindings' system */
    nxp_op_bindings_init();
}



/**
 * nxp_elements_destroy:
 * 
 * Finalize elements subsystem.
 **/
void nxp_elements_destroy(void)
{
#ifdef NXP_ELEMENT_DEBUG
    GSList *l;
#endif
/*      pthread_setspecific(nxp_element_context_tsd_key, NULL); */
#ifdef NXP_ELEMENT_DEBUG
    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	printf("identifiers still in memory: %i\n", nxp_identifier_global_count);
	printf("element references still assigned: %i\n", nxp_element_references_count);
	printf("elements still in memory: %i\n", nxp_element_global_count);
	printf("------------ NxpElement's in memory follow ------------\n");
	for (l = elements_list; l; l = l->next)
	    nxp_elem_printer(NXP_ELEMENT(l->data));
	printf("------------ End of elements in memory ------------\n");
	g_slist_free(elements_list);
    }
    /* Finalize operator bindings */
    nxp_op_bindings_fini();

    /* Finalize element types */
    nxp_element_type_fini();

#endif
}



gboolean nxp_element_is_true(NxpElement *el, NxpEvalContext *context, GError **error)
{
    NxpElementTypeInfo *typeinfo;
    GError *error1 = NULL;

    g_return_val_if_fail(el, FALSE);
    if (nxp_element_check_thread_abortion_request()) return FALSE;

    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(el);
    g_return_val_if_fail(typeinfo, FALSE);
    if (typeinfo->truth_func) {
	gboolean ret;
	ret = typeinfo->truth_func(el, context, &error1);
	if (error1) g_propagate_error(error, error1);
	return ret;
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_NOT_IMPLEMENTED,
		    _("Don't know how to evaluate element of type '%s' as truth value"), NXP_ELEMENT_GET_TYPE_NAME(el));
    }
    return FALSE;
}
