/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

/* Complex R relation */

#include <math.h>

#include "int.h"
#include "tree.h"
#include "nxp_complex.h"
#include "errors.h"
#include "i18n.h"

#include "ourlibc.h"

#include "op_bindings.h"
#include "types_misc.h"

#define BIND_SYM(eval,simp,op)    nxp_bind_operator_function(op,NXP_ELEMENT_TYPE_COMPLEX, NXP_ELEMENT_TYPE_R, \
   			                              eval, simp,  NXP_OP_BIND_SYMMETRIC, NULL);
#define BIND_C_R(eval, simp, op) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_COMPLEX, \
                                                             NXP_ELEMENT_TYPE_R, \
   			                                     eval, simp,  0, NULL);
#define BIND_R_C(eval,simp,op) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_R, \
                                                             NXP_ELEMENT_TYPE_COMPLEX, \
   			                                     eval, simp,  0, NULL);


/* SYMMETRIC OPERATORS */

static NxpElement* atrib_error(NxpOp op, NxpElement *a, NxpElement *b, 
			       NxpEvalContext *context, GError **error)
{
  /* TODO: define a non lvalue GError and use it here */
  /* So far, not needed because parser gives the error, already! */
  return NULL;
}

static NxpElement* add(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    NxpElement *cpx = nxp_element_complex_new(NULL, NULL);

    NXP_COMPLEX_REAL(cpx) = nxp_element_tree_simplify_op(op,
							 NXP_COMPLEX_REAL(a),
							 b, context, error);
    _element_vector_set(NXP_ELEMENT_COMPLEX(cpx)->vector, 1,
			_element_vector_index(NXP_ELEMENT_COMPLEX(a)->vector, 1));
    return cpx;
}

static NxpElement* add_eval(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    NxpElement *cpx = nxp_element_complex_new(NULL, NULL);

    NXP_COMPLEX_REAL(cpx) = nxp_element_tree_eval_op(op,
						     NXP_COMPLEX_REAL(a),
						     b, context, error);
    _element_vector_set(NXP_ELEMENT_COMPLEX(cpx)->vector, 1,
			_element_vector_index(NXP_ELEMENT_COMPLEX(a)->vector, 1));
    return cpx;
}

static NxpElement* mul(NxpOp op, NxpElement *a, NxpElement *b, 
			     NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_R(b), NULL);

    NxpElement *cpx;
    NxpElement *vector;

    if (nxp_element_is_true(b, context, error)) {
	vector = nxp_element_tree_simplify_op(NXP_OP_STAR,
				    NXP_ELEMENT(NXP_ELEMENT_COMPLEX(a)->vector),
				    b, context, error);
	cpx = nxp_element_complex_new_from_vector(vector);
	NXP_ELEMENT_UNREF(vector);
    }
    else
	cpx = NXP_ELEMENT_REF(b);
    return cpx;
}

static NxpElement* mul_eval(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    NxpElement *cpx;
    NxpElement *vector;

    if (nxp_element_is_true(b, context, error)) {
	vector = nxp_element_tree_eval_op(NXP_OP_STAR,
				    NXP_ELEMENT(NXP_ELEMENT_COMPLEX(a)->vector),
				    b, context, error);
	cpx = nxp_element_complex_new_from_vector(vector);
	NXP_ELEMENT_UNREF(vector);
    }
    else
	cpx = NXP_ELEMENT_REF(b);
    return cpx;
}

//---- ASYMS
static NxpElement* sub_r_c(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    NxpElementComplex *cpx = (NxpElementComplex *)nxp_element_complex_new(NULL, NULL);

    NXP_COMPLEX_REAL(cpx) = nxp_element_tree_simplify_op(op,
					a, NXP_COMPLEX_REAL(b), context, error);
    NXP_COMPLEX_IMG(cpx) = nxp_element_r_opposed(
					NXP_ELEMENT_R(NXP_COMPLEX_IMG(b)));
    return NXP_ELEMENT(cpx);
}

static NxpElement* sub_r_c_eval(NxpOp op, NxpElement *a, NxpElement *b,
				NxpEvalContext *context, GError **error)
{
    NxpElementComplex *cpx = (NxpElementComplex *)nxp_element_complex_new(NULL, NULL);

    NXP_COMPLEX_REAL(cpx) = nxp_element_tree_eval_op(op,
					a, NXP_COMPLEX_REAL(b), context, error);
    NXP_COMPLEX_IMG(cpx) = nxp_element_r_opposed(
					NXP_ELEMENT_R(NXP_COMPLEX_IMG(b)));
    return NXP_ELEMENT(cpx);
}

static NxpElement* div_c_r(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    NxpElement *cpx;

    if (nxp_element_is_true(b, context, error)) {
	if (nxp_element_is_true(a, context, error)) {
	    NxpElement *vector;

	    vector = nxp_element_tree_simplify_op(op,
				    NXP_ELEMENT(NXP_ELEMENT_COMPLEX(a)->vector),
				    b, context, error);
	    cpx = nxp_element_complex_new_from_vector(vector);
	    NXP_ELEMENT_UNREF(vector);
	}
	else
	    cpx = nxp_element_int_new(0);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	return NULL;
    }
    return cpx;
}

static NxpElement* div_c_r_eval(NxpOp op, NxpElement *a, NxpElement *b,
				NxpEvalContext *context, GError **error)
{
    NxpElement *cpx;

    if (nxp_element_is_true(b, context, error)) {
	if (nxp_element_is_true(a, context, error)) {
	    NxpElement *vector;

	    vector = nxp_element_tree_eval_op(op,
				NXP_ELEMENT(NXP_ELEMENT_COMPLEX(a)->vector),
				b, context, error);
	    cpx = nxp_element_complex_new_from_vector(vector);
	    NXP_ELEMENT_UNREF(vector);
	}
	else
	    cpx = nxp_element_int_new(0);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	return NULL;
    }
    return cpx;
}
static NxpElement* div_r_c(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error)
{
    NxpElement *aa, *bb, *mb, *den;
    NxpElement *cpx;

    if (nxp_element_is_true(b, context, error)) {
	if (nxp_element_is_true(a, context, error)) {
	    aa = nxp_element_tree_simplify_op(NXP_OP_STAR,
		    NXP_COMPLEX_REAL(b), NXP_COMPLEX_REAL(b), context, error);

	    bb = nxp_element_tree_simplify_op(NXP_OP_STAR,
		    NXP_COMPLEX_IMG(b), NXP_COMPLEX_IMG(b), context, error);

	    den = nxp_element_tree_simplify_op(NXP_OP_PLUS, aa, bb,
					       context, error);
	    NXP_ELEMENT_UNREF(aa);
	    NXP_ELEMENT_UNREF(bb);

	    aa = nxp_element_tree_simplify_op(NXP_OP_STAR,
		    NXP_COMPLEX_REAL(b), a, context, error);

	    mb = nxp_element_r_opposed(NXP_ELEMENT_R(NXP_COMPLEX_IMG(b)));
	    bb = nxp_element_tree_simplify_op(NXP_OP_STAR,
					      mb, a, context, error);
	    NXP_ELEMENT_UNREF(mb);

	    cpx = nxp_element_complex_new(NULL, NULL);
	    NXP_COMPLEX_REAL(cpx) = nxp_element_tree_simplify_op(
							NXP_OP_SLASH, aa, den,
							context, error);
	    NXP_COMPLEX_IMG(cpx) = nxp_element_tree_simplify_op(
							NXP_OP_SLASH, bb, den,
							context, error);

	    NXP_ELEMENT_UNREF(aa);
	    NXP_ELEMENT_UNREF(bb);
	    NXP_ELEMENT_UNREF(den);

	    return cpx;
	} else
	    return nxp_element_int_new(0);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	return NULL;
    }
}

static NxpElement* div_r_c_eval(NxpOp op, NxpElement *a, NxpElement *b, 
				NxpEvalContext *context, GError **error)
{
    NxpElement *aa, *bb, *den;
    NxpElement *cpx;

    if (nxp_element_is_true(b, context, error)) {
	if (nxp_element_is_true(a, context, error)) {
	    aa = nxp_element_tree_eval_op(NXP_OP_STAR,
		    _element_vector_index(NXP_ELEMENT_COMPLEX(b)->vector, 0),
		    _element_vector_index(NXP_ELEMENT_COMPLEX(b)->vector, 0),
		    context, error);
	    bb = nxp_element_tree_eval_op(NXP_OP_STAR,
		    _element_vector_index(NXP_ELEMENT_COMPLEX(b)->vector, 1),
		    _element_vector_index(NXP_ELEMENT_COMPLEX(b)->vector, 1),
		    context, error);
	    den = nxp_element_tree_eval_op(NXP_OP_PLUS, aa, bb,
					       context, error);
	    NXP_ELEMENT_UNREF(aa);
	    NXP_ELEMENT_UNREF(bb);

	    aa = nxp_element_tree_eval_op(NXP_OP_STAR,
		    _element_vector_index(NXP_ELEMENT_COMPLEX(b)->vector, 0),
		    a, context, error);

	    bb = nxp_element_r_opposed(
		    NXP_ELEMENT_R(NXP_COMPLEX_IMG(b)));

	    cpx = nxp_element_complex_new(NULL, NULL);
	    _element_vector_index(NXP_ELEMENT_COMPLEX(cpx)->vector, 0)
		= nxp_element_tree_eval_op(NXP_OP_SLASH, aa, den,
					       context, error);
	    _element_vector_index(NXP_ELEMENT_COMPLEX(cpx)->vector, 1)
		= nxp_element_tree_eval_op(NXP_OP_SLASH, bb, den,
					       context, error);

	    NXP_ELEMENT_UNREF(aa);
	    NXP_ELEMENT_UNREF(bb);
	    NXP_ELEMENT_UNREF(den);

	    return cpx;
	} else
	    return nxp_element_int_new(0);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	return NULL;
    }
}

static NxpElement* hat_r_c(NxpOp op, NxpElement *a, NxpElement *b,
				NxpEvalContext *context, GError **error)
{
    NxpElement *new_b;
    if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(NXP_COMPLEX_IMG(NXP_ELEMENT_COMPLEX(b))), 0) == 0) {
	new_b = NXP_COMPLEX_REAL(NXP_ELEMENT_COMPLEX(b));
	return nxp_element_tree_simplify_op(op, a, new_b, context, error);
    }
    else
	return nxp_element_tree_new(op, a, b);
}

static NxpElement* hat_r_c_eval(NxpOp op, NxpElement *a, NxpElement *b,
				NxpEvalContext *context, GError **error)
{
    return nxp_element_complex_new_from_dbl(
	        cpowl((nxp_complex_t)nxp_element_r_get_value_as_double(
						    NXP_ELEMENT_R(a)),
		      nxp_element_complex_get_value(NXP_ELEMENT_COMPLEX(b))));
}

static NxpElement* hat_c_r(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    NxpElement *cpx, *cpx1, *res;
    int pow;

    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_R(b), NULL);

    if (nxp_element_r_is_integer(NXP_ELEMENT_R(b)) && nxp_element_r_compare_int(NXP_ELEMENT_R(b), 0) >= 0) {

	pow = nxp_element_r_get_value_as_int(NXP_ELEMENT_R(b));
	cpx = NXP_ELEMENT_REF(a);
	res = nxp_element_int_new(1);
	while(pow != 0) {
	    if (pow & 1) {
		cpx1 = res;
		res = nxp_element_tree_simplify_op(NXP_OP_STAR, cpx, cpx1,
						   context, error);
		NXP_ELEMENT_UNREF(cpx1);
	    }
	    cpx1 = nxp_element_tree_simplify_op(NXP_OP_STAR, cpx, cpx,
						context, error);
	    NXP_ELEMENT_UNREF(cpx);
	    cpx = cpx1;
	    pow >>= 1;    
	}
	NXP_ELEMENT_UNREF(cpx);
    }
    else
	return nxp_element_complex_new_from_dbl(cpowl(nxp_element_complex_get_value(NXP_ELEMENT_COMPLEX(a)),
					 (nxp_complex_t)nxp_element_r_get_value_as_double(NXP_ELEMENT_R(b))));
    return res;
}

static NxpElement* hat_c_r_eval(NxpOp op, NxpElement *a, NxpElement *b,
				NxpEvalContext *context, GError **error)
{
    NxpElement *cpx, *cpx1, *res;
    int pow;

    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_R(b), NULL);

    if (nxp_element_r_is_integer(NXP_ELEMENT_R(b)) && nxp_element_r_compare_int(NXP_ELEMENT_R(b), 0) >= 0) {

	pow = nxp_element_r_get_value_as_int(NXP_ELEMENT_R(b));
	cpx = NXP_ELEMENT_REF(a);
	res = nxp_element_int_new(1);
	while(pow != 0) {
	    if (pow & 1) {
		cpx1 = res;
		res = nxp_element_tree_eval_op(NXP_OP_STAR, cpx, cpx1,
						   context, error);
		NXP_ELEMENT_UNREF(cpx1);
	    }
	    cpx1 = nxp_element_tree_eval_op(NXP_OP_STAR, cpx, cpx,
						context, error);
	    NXP_ELEMENT_UNREF(cpx);
	    cpx = cpx1;
	    pow >>= 1;    
	}
	NXP_ELEMENT_UNREF(cpx);
    }
    else
	return nxp_element_complex_new_from_dbl(cpowl(nxp_element_complex_get_value(NXP_ELEMENT_COMPLEX(a)),
					 (nxp_complex_t)nxp_element_r_get_value_as_double(NXP_ELEMENT_R(b))));
    return res;
}
//// AND ////
static NxpElement* nxp_complex_and_r(NxpOp op, NxpElement *a, NxpElement *b,
				       NxpEvalContext *context, GError **error)
{
    nxp_complex_t aa, bb;
    aa = nxp_element_complex_get_value(NXP_ELEMENT_COMPLEX(a));
    bb = (nxp_complex_t)nxp_element_r_get_value_as_double(NXP_ELEMENT_R(b));
    return nxp_element_complex_new_from_dbl(aa?bb:aa);
}

//// OR ////
static NxpElement* nxp_complex_or_r(NxpOp op, NxpElement *a, NxpElement *b,
				    NxpEvalContext *context, GError **error)
{
    nxp_complex_t aa, bb;
    aa = nxp_element_complex_get_value(NXP_ELEMENT_COMPLEX(a));
    bb = (nxp_complex_t)nxp_element_r_get_value_as_double(NXP_ELEMENT_R(b));
    return nxp_element_complex_new_from_dbl(aa?aa:bb);
}

static NxpElement* eq_c_r(NxpOp op, NxpElement *a, NxpElement *b,
			  NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_R(b), NULL);

    if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(a), 0) == 0
	    && nxp_element_eq(NXP_COMPLEX_REAL(a), b, context, error))
	return nxp_element_int_new(1);
    else
	return nxp_element_int_new(0);
}

static NxpElement* ne_c_r(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    g_return_val_if_fail(NXP_ELEMENT_IS_COMPLEX(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_R(b), NULL);

    if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(a), 0) != 0
	    || !nxp_element_eq(NXP_COMPLEX_REAL(a), b, context, error))
	return nxp_element_int_new(1);
    else
	return nxp_element_int_new(0);
}

void nxp_element_complex_r_operator_init()
{
    BIND_SYM(add_eval, add, NXP_OP_PLUS);
    BIND_SYM(mul_eval, mul, NXP_OP_STAR);

    BIND_C_R(add_eval, add, NXP_OP_MINUS);
    BIND_R_C(sub_r_c_eval, sub_r_c, NXP_OP_MINUS);

    BIND_C_R(div_c_r_eval, div_c_r, NXP_OP_SLASH);
    BIND_R_C(div_r_c_eval, div_r_c, NXP_OP_SLASH);

    BIND_C_R(hat_c_r_eval, hat_c_r, NXP_OP_HAT);
    BIND_R_C(hat_r_c_eval, hat_r_c, NXP_OP_HAT);

    BIND_SYM(eq_c_r, eq_c_r, NXP_OP_EQ2);
    BIND_SYM(ne_c_r, eq_c_r, NXP_OP_EXCL_EQ);
  /* NXP_OP_DOT3 --> This case is handled under complex.c or int.c */
  /* NXP_OP_PLUS2_LEFT --> This case is handled under complex.c or int.c */
  /* NXP_OP_PLUS2_RIGHT --> This case is handled under complex.c or int.c */
  /* NXP_OP_MINUS2_LEFT --> This case is handled under complex.c or int.c */
  /* NXP_OP_MINUS2_RIGHT --> This case is handled under complex.c or int.c */

  /* To investigate
     ---------------------------------------------------------------
     BIND_COMPLEX_INT(nxp_complex_gt_int, NXP_OP_GT);
     BIND_INT_COMPLEX(nxp_int_gt_complex, NXP_OP_GT);
     BIND_COMPLEX_INT(nxp_complex_ge_int, NXP_OP_GT_EQ);
     BIND_INT_COMPLEX(nxp_int_ge_complex, NXP_OP_GT_EQ);
     BIND_COMPLEX_INT(nxp_complex_lt_int, NXP_OP_LT);
     BIND_INT_COMPLEX(nxp_int_lt_complex, NXP_OP_LT);
     BIND_COMPLEX_INT(nxp_complex_le_int, NXP_OP_LT_EQ);
     BIND_INT_COMPLEX(nxp_int_le_complex, NXP_OP_LT_EQ); */

    BIND_SYM(nxp_complex_and_r, nxp_complex_and_r, NXP_OP_SLASH_BSLASH);
    BIND_SYM(nxp_complex_or_r, nxp_complex_or_r, NXP_OP_BSLASH_SLASH);

  /* NXP_OP_TILDE_LEFT --> This case is handled under complex.c or int.c */
  /* NXP_OP_EXCL_RIGHT --> This case is handled under complex.c or int.c */

  BIND_SYM(atrib_error, atrib_error, NXP_OP_EQ);
  BIND_SYM(atrib_error, atrib_error, NXP_OP_COLON_EQ);
}
