/* -*- Mode: C; c-file-style: "stroustrup"; coding: latin-1-unix -*- */
#include <stdio.h>
#include "container.h"

#include "message.h"
#include "errors.h"
#include "tree.h"
#include "int.h"
#include "vector.h"
#include "matrix.h"
#include "variable.h"
#include <variables.h>
#include "element_output.h"
#include "types_misc.h"
#include "void.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

/**
 * nxp_element_container_new: -create container element
 * @size: the number of elements in the container
 * @data: an array of pointers to each element of the container; this block of
 *        memory should have a size of at least (and, if possible, no more than) 
 *        @size*sizeof(&NxpElement*); if %NULL, a new array of pointers (initialized to %NULL)
 *        will be created to hold @size elements; Note: if @data is non null, the container element
 *        will take ownership (keep) of this array.
 * 
 * Create a new container element.
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_container_new(NxpElement *parent, NxpElement *index)
{
    NxpElement *elem;

    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementContainer));
    NXP_ELEMENT_CONTAINER(elem)->parent = NXP_ELEMENT_REF(parent);
    NXP_ELEMENT_CONTAINER(elem)->index = NXP_ELEMENT_REF(index);

    return elem;
}

static void nxp_destroy_container(NxpElement *elem)
{
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(elem);
    NXP_ELEMENT_UNREF(ctn->parent);
    NXP_ELEMENT_UNREF(ctn->index);
}

static NxpElement** nxp_container_get_pointer(NxpElement *parent, NxpElement *index, NxpEvalContext *context, GError **error)
{
    GError *err = NULL;

    if (NXP_ELEMENT_IS_VARIABLE(parent)) {
	NxpElement *value;
	value = nxp_variable_get(NXP_ELEMENT_VARIABLE(parent)->ident, context);
	return nxp_container_get_pointer(value, index, context, error);
    }
    else if (NXP_ELEMENT_IS_VECTOR(parent)) {
	gint i;
	if (!NXP_ELEMENT_IS_INT(index)) {
	    index = nxp_element_eval(index, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		return NULL;
	    }
	    if (!NXP_ELEMENT_IS_INT(index)) {
		g_set_error(error, nxp_error_domain, NXP_ERROR_UNDEFINED, _("You ask for an index that is not an integer"));
		NXP_ELEMENT_UNREF(index);
		return NULL;
	    }
	}
	i = nxp_element_int_get_si(index, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	if (i >= 1 && i <= NXP_ELEMENT_VECTOR(parent)->size)
	    return &_element_vector_index(NXP_ELEMENT_VECTOR(parent), i - 1);
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_OUT_OF_RANGE, _("You are trying to access out of the vector"));
	    return NULL;
	}
    }
    else if (NXP_ELEMENT_IS_CONTAINER(parent)) {
	NxpElement **pointer;
	GError *err = NULL;

	pointer = nxp_container_get_pointer(NXP_ELEMENT_CONTAINER(parent)->parent, NXP_ELEMENT_CONTAINER(parent)->index, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	return nxp_container_get_pointer(*pointer, index, context, error);
    }
    else if (NXP_ELEMENT_IS_MATRIX(parent)) {
	NxpElementVector *vector;
	NxpElementMatrix *m = NXP_ELEMENT_MATRIX(parent);
	gint w = m->width;
	gint i, j;

	i = nxp_element_int_get_si(index, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	if (i >= 1 && i <= m->height) {
	    vector = (NxpElementVector *)nxp_element_vector_new(w, NULL);
	    for (j = 0; j < w; j++)
		_element_vector_index(vector, j) = NXP_ELEMENT_REF(_element_matrix_index(m, i, j));
	    return (NxpElement **)&vector; /* FIXME: Memory leak !!! */
	}
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_OUT_OF_RANGE, _("You are trying to access out of the matrix"));
	    return NULL;
	}
    }

    nxp_elem_printer(parent);
    nxp_elem_printer(index);
    g_assert_not_reached();
}

static NxpElement* eval_assign(NxpOp op, NxpElement *a, NxpElement *b,
			       NxpEvalContext *context, GError **error)
{
    NxpElement **el;
    NxpElement *value;
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(a);
    GError *err = NULL;

    value = nxp_element_eval(b, context, &err);
    if (err) {
	NXP_ELEMENT_XUNREF(value);
	g_propagate_error(error, err);
	return NULL;
    }
    el = nxp_container_get_pointer(ctn->parent, ctn->index, context, &err);
    if (err) {
	NXP_ELEMENT_UNREF(value);
	g_propagate_error(error, err);
	return NULL;
    }
    NXP_ELEMENT_UNREF(*el);
    *el = value;
    return NXP_ELEMENT_REF(a);
}

static NxpElement* simp_assign(NxpOp op, NxpElement *a, NxpElement *b,
			       NxpEvalContext *context, GError **error)
{
    NxpElement **el;
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(a);
    NxpElement *value;
    GError *err = NULL;

    value = nxp_element_simplify(b, context, &err);
    if (err) {
	NXP_ELEMENT_XUNREF(value);
	g_propagate_error(error, err);
	return NULL;
    }
    el = nxp_container_get_pointer(ctn->parent, ctn->index, context, &err);
    if (err) {
	NXP_ELEMENT_UNREF(value);
	g_propagate_error(error, err);
	return NULL;
    }
    NXP_ELEMENT_UNREF(*el);
    *el = value;
    return NXP_ELEMENT_REF(a);
}

static NxpElement* eval_store(NxpOp op, NxpElement *a, NxpElement *b,
			       NxpEvalContext *context, GError **error)
{
    NxpElement **el;
    NxpElement *value;
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(a);
    GError *err = NULL;

    value = NXP_ELEMENT_REF(b);
    if (err) {
	NXP_ELEMENT_XUNREF(value);
	g_propagate_error(error, err);
	return NULL;
    }
    el = nxp_container_get_pointer(ctn->parent, ctn->index, context, &err);
    if (err) {
	NXP_ELEMENT_UNREF(value);
	g_propagate_error(error, err);
	return NULL;
    }
    NXP_ELEMENT_UNREF(*el);
    *el = value;
    return NXP_ELEMENT_REF(a);
}

#define BIND(op)    nxp_bind_operator_function(op,	\
					       NXP_ELEMENT_TYPE_CONTAINER, \
					       NXP_ELEMENT_TYPE_CONTAINER, \
    			                       map_op_eval, map_op_simplify,	\
					       0, NULL);

#define BIND_FUNC(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_CONTAINER, \
							 NXP_ELEMENT_TYPE_CONTAINER, \
    			                                 eval, symp, 0, NULL);
#define BIND_LEFT(eval,symp,op)    nxp_bind_operator_function(op, 0,         \
                                                         NXP_ELEMENT_TYPE_CONTAINER, \
   			                                 eval, symp, 0, NULL);

NxpElement* nxp_eval_container(NxpElement *elem, NxpEvalContext *context, GError **error) {
    NxpElement **el;
    GError *err = NULL;
    NxpElementContainer* ctn = NXP_ELEMENT_CONTAINER(elem);

    el = nxp_container_get_pointer(ctn->parent, ctn->index, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    return nxp_element_eval(*el, context, error);
}

NxpElement* nxp_simplify_container(NxpElement *elem, NxpEvalContext *context, GError **error) {
    NxpElement **el;
    GError *err = NULL;
    NxpElementContainer* ctn = NXP_ELEMENT_CONTAINER(elem);
    el = nxp_container_get_pointer(ctn->parent, ctn->index, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    return nxp_element_simplify(*el, context, error);
}

static GString* repr_func(NxpElement *elem, int *precedence)
{
    NxpElementTypeInfo *typeinfo;
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(elem);
    GString *str;
    GString *str1;
    int prec;

    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(ctn->parent);
    str = typeinfo->repr_func(ctn->parent, &prec);
    if (prec > 10) {
	str = g_string_prepend(str, "(");
	str = g_string_append(str, ")");
    }

    typeinfo = NXP_ELEMENT_GET_TYPE_INFO(ctn->index);
    str1 = typeinfo->repr_func(ctn->index, &prec);
    if (prec > 10) {
	str1 = g_string_prepend(str1, "(");
	str1 = g_string_append(str1, ")");
    }
    g_string_append(str, " -> ");
    g_string_append(str, str1->str);
    g_string_free(str1, TRUE);

    if (precedence)
	*precedence = 10;
    return str;
}

/*static void
iterator_func(NxpElement *elem,
              NxpElementVisitorFunc func,
              gpointer user_data,
              NxpEvalContext *context)
{
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(elem);
    NxpElement *el;
    gint i;

    i = ctn->index;
    el = _element_vector_index(NXP_ELEMENT_VECTOR(ctn->parent), i);

    func(el, user_data);
    if (el->type_info->iterator_func)
	el->type_info->iterator_func(el, func, user_data, context);
}
*/

static gboolean dependson_func(NxpElement *elem, NxpElement *variable, NxpEvalContext *context, GError **error)
{
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(elem);
    gboolean retval = FALSE;
    GError *err = NULL;

    retval = nxp_element_depends_on(ctn->parent, variable, context, &err);
    if (err) {
	g_propagate_error(error, err);
	return FALSE;
    }
    if (!retval) {
	retval = nxp_element_depends_on(ctn->index, variable, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return FALSE;
	}
    }
    return retval;
}

static NxpElement* substitute_func(NxpElement *self, NxpElement *var, NxpElement *el, NxpEvalContext *context, GError **error)
{
    NxpElement *parent, *index, *retval;
    NxpElementContainer *ctn = NXP_ELEMENT_CONTAINER(self);
    GError *err = NULL;

    parent = nxp_element_substitute(ctn->parent, var, el, context, &err);
    if (err) {
	NXP_ELEMENT_UNREF(NXP_ELEMENT(retval));
	g_propagate_error(error, err);
	return NULL;
    }
    index = nxp_element_substitute(ctn->index, var, el, context, &err);
    if (err) {
	NXP_ELEMENT_UNREF(NXP_ELEMENT(retval));
	g_propagate_error(error, err);
	return NULL;
    }
    retval = nxp_element_container_new(parent, index);
    NXP_ELEMENT_UNREF(parent);
    NXP_ELEMENT_UNREF(index);
    return NXP_ELEMENT(retval);
}

int nxp_element_container_get_type(void)
{
    static int nxp_element_type_container = 0;

    if (nxp_element_type_container == 0) {
	nxp_element_type_container = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_container,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_container;
}

void nxp_element_container_type_init() 
{
    typeinfo = nxp_element_type_get_info(nxp_element_container_get_type());
    typeinfo->name = g_strdup("container");
    typeinfo->destroy_func = nxp_destroy_container;
    typeinfo->eval_func = nxp_eval_container;
    typeinfo->simplify_func = nxp_simplify_container;
    typeinfo->derive_func = NULL;
    typeinfo->str_func = NULL;
    typeinfo->repr_func = repr_func;
    typeinfo->truth_func = NULL;
    typeinfo->iterator_func = NULL;
    typeinfo->dependson_func = dependson_func;
    typeinfo->substitute_func = substitute_func;

    nxp_bind_operator_function(NXP_OP_EQ, NXP_ELEMENT_TYPE_CONTAINER, NXP_ELEMENT_TYPE_ANY,
			       eval_assign, simp_assign, 
			       0, NULL);
    nxp_bind_operator_function(NXP_OP_COLON_EQ, NXP_ELEMENT_TYPE_CONTAINER, NXP_ELEMENT_TYPE_ANY,
			       eval_store, eval_store, 
			       0, NULL);
}
