/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include <math.h>

#include "int.h"
#include "real.h"
#include "frac.h"
#include "tree.h"
#include "op_bindings.h"
#include "types_misc.h"
#include "errors.h"
#include "i18n.h"

static NxpElement *simp_add(NxpOp op, NxpElement* a, NxpElement *b,
			    NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;
    mpz_t n;

    mpz_init(n);
    mpz_mul(n, mpq_denref(NXP_ELEMENT_FRAC(a)->value), NXP_ELEMENT_INT(b)->value);

    mpq_init(i);
    mpz_add(mpq_numref(i), mpq_numref(NXP_ELEMENT_FRAC(a)->value), n);
    mpz_set(mpq_denref(i), mpq_denref(NXP_ELEMENT_FRAC(a)->value));
    el = nxp_element_frac_new_from_mpq(i);
    mpz_clear(n);
    return el;
}

static NxpElement *simp_f_i_sub(NxpOp op, NxpElement* a, NxpElement *b,
				NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;
    mpz_t n;

    mpz_init(n);
    mpz_mul(n, mpq_denref(NXP_ELEMENT_FRAC(a)->value), NXP_ELEMENT_INT(b)->value);

    mpq_init(i);
    mpz_sub(mpq_numref(i), mpq_numref(NXP_ELEMENT_FRAC(a)->value), n);
    mpz_set(mpq_denref(i), mpq_denref(NXP_ELEMENT_FRAC(a)->value));
    el = nxp_element_frac_new_from_mpq(i);
    mpz_clear(n);
    return el;
}

static NxpElement *simp_i_f_sub(NxpOp op, NxpElement* a, NxpElement *b,
				NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;
    mpz_t n;

    mpz_init(n);
    mpz_mul(n, mpq_denref(NXP_ELEMENT_FRAC(b)->value), NXP_ELEMENT_INT(a)->value);

    mpq_init(i);
    mpz_sub(mpq_numref(i), n, mpq_numref(NXP_ELEMENT_FRAC(b)->value));
    mpz_set(mpq_denref(i), mpq_denref(NXP_ELEMENT_FRAC(b)->value));
    el = nxp_element_frac_new_from_mpq(i);
    mpz_clear(n);
    return el;
}

static NxpElement *simp_mul(NxpOp op, NxpElement* a, NxpElement *b,
			    NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;
    mpq_init(i);
    mpz_mul(mpq_numref(i), mpq_numref(NXP_ELEMENT_FRAC(a)->value), NXP_ELEMENT_INT(b)->value);
    mpz_set(mpq_denref(i), mpq_denref(NXP_ELEMENT_FRAC(a)->value));
    mpq_canonicalize(i);
    if (mpz_cmp_ui(mpq_denref(i), 1))
	el = nxp_element_frac_new_from_mpq(i);
    else {
	mpz_t z;
	mpz_init_set(z, mpq_numref(i));
	mpq_clear(i);
	el = nxp_element_int_new_from_mpz(z);
    }
    return el;
}

/*static NxpElement *simp_eq2(NxpOp op, NxpElement* a, NxpElement* b,
			    NxpEvalContext *context, GError **error) {
    int sgn = NXP_ELEMENT_FRAC_POSITIVE(a) ? 1 : -1;
    g_return_val_if_fail(NXP_ELEMENT_IS_FRAC(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_INT(b), NULL);
    return nxp_element_int_new(
	sgn * NXP_ELEMENT_FRAC_NUM(a) -
	NXP_ELEMENT_INT(b)->value * NXP_ELEMENT_FRAC_DENUM(a) == 0 ? (op == NXP_OP_EQ2) : (op != NXP_OP_EQ2));
}*/

static NxpElement *simp_f_i_div(NxpOp op, NxpElement* a, NxpElement *b,
				NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;

    if (mpz_cmp_si(NXP_ELEMENT_INT(b)->value, 0)) {
	mpq_init(i);
	mpz_set(mpq_numref(i), mpq_numref(NXP_ELEMENT_FRAC(a)->value));
	mpz_mul(mpq_denref(i), mpq_denref(NXP_ELEMENT_FRAC(a)->value),
		NXP_ELEMENT_INT(b)->value);
	mpq_canonicalize(i);
	el = nxp_element_frac_new_from_mpq(i);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
			_("divide by zero"));
	return NULL;
    }
    return el;
}

static NxpElement *simp_i_f_div(NxpOp op, NxpElement* a, NxpElement *b,
				NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;

    if (mpz_cmp_si(mpq_numref(NXP_ELEMENT_FRAC(b)->value), 0)) {
	mpq_init(i);
	mpz_mul(mpq_numref(i), mpq_denref(NXP_ELEMENT_FRAC(b)->value),
		NXP_ELEMENT_INT(a)->value);
	mpz_set(mpq_denref(i), mpq_numref(NXP_ELEMENT_FRAC(b)->value));
	el = nxp_element_frac_new_from_mpq(i);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
			_("divide by zero"));
	return NULL;
    }
    return el;
}

static NxpElement *simp_f_i_hat(NxpOp op, NxpElement* a, NxpElement* b,
				NxpEvalContext *context, GError **error)
{
    mpq_t q;
    nxp_int_t pow = mpz_get_si(NXP_ELEMENT_INT(b)->value);
    NxpElement *el;
    int p = (pow >= 0)?pow:-pow;

    mpq_init(q);
    mpz_pow_ui(mpq_numref(q), mpq_numref(NXP_ELEMENT_FRAC(a)->value), p);
    mpz_pow_ui(mpq_denref(q), mpq_denref(NXP_ELEMENT_FRAC(a)->value), p);
    if (pow < 0) {
	if (mpz_sgn(mpq_numref(q)) != 0) {
	    mpq_inv(q, q);
	    el = nxp_element_frac_new_from_mpq(q);
	}
	else {
	    el = NULL;
	    mpq_clear(q);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
		    _("divide by zero"));
	}
    }
    else
	el = nxp_element_frac_new_from_mpq(q);
    return el;
    
}

static NxpElement *nequals(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    if (mpz_cmp_ui(mpq_denref(NXP_ELEMENT_FRAC(a)->value), 1)
	    || mpz_cmp(mpq_numref(NXP_ELEMENT_FRAC(a)->value), NXP_ELEMENT_INT(b)->value))
	return nxp_element_int_new(1);
    else
	return nxp_element_int_new(0);
}

#define BIND_F_I(op,simp,eval) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_FRAC, NXP_ELEMENT_TYPE_INT, \
   			                              eval, simp,  0, NULL);
#define BIND_I_F(op,simp) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_INT, NXP_ELEMENT_TYPE_FRAC, \
   			                              NULL, simp,  0, NULL);

#define BIND_SYM(op,simp) nxp_bind_operator_function(op, NXP_ELEMENT_TYPE_FRAC, \
							     NXP_ELEMENT_TYPE_INT, \
   			                              NULL, simp,  NXP_OP_BIND_SYMMETRIC, NULL);


void nxp_element_frac_int_operator_init(void)
{
    BIND_SYM(NXP_OP_PLUS, simp_add);
    BIND_SYM(NXP_OP_STAR, simp_mul);

    BIND_F_I(NXP_OP_MINUS, simp_f_i_sub, NULL);
    BIND_I_F(NXP_OP_MINUS, simp_i_f_sub);

    BIND_F_I(NXP_OP_SLASH, simp_f_i_div, NULL);
    BIND_I_F(NXP_OP_SLASH, simp_i_f_div);

    BIND_F_I(NXP_OP_HAT, simp_f_i_hat, NULL);
//    BIND_SYM(NXP_OP_EQ2, simp_eq2);
//    BIND_SYM(NXP_OP_EXCL_EQ, simp_eq2);

/*   BIND_I_R(hat_int_real, NXP_OP_HAT); */

/*   BIND_R_I(gt_real_int, NXP_OP_GT); */
/*   BIND_I_R(gt_int_real, NXP_OP_GT); */

/*   BIND_R_I(lt_real_int, NXP_OP_LT); */
/*   BIND_I_R(lt_int_real, NXP_OP_LT); */

/*   BIND_R_I(gte_real_int, NXP_OP_GT_EQ); */
/*   BIND_I_R(gte_int_real, NXP_OP_GT_EQ); */

/*   BIND_R_I(lte_real_int, NXP_OP_LT_EQ); */
/*   BIND_I_R(lte_int_real, NXP_OP_LT_EQ); */

   BIND_SYM(NXP_OP_EXCL_EQ, nequals);

/*   BIND_R_I(and_real_int, NXP_OP_SLASH_BSLASH); */
/*   BIND_I_R(and_int_real, NXP_OP_SLASH_BSLASH); */

/*   BIND_R_I(or_real_int, NXP_OP_BSLASH_SLASH); */
/*   BIND_I_R(or_int_real, NXP_OP_BSLASH_SLASH); */

}
