/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include <stdio.h>
#include <math.h>
#include "int.h"
#include "real.h"
#include "frac.h"
#include "tree.h"
#include "vector.h"
#include "message.h"
#include "op_bindings.h"
#include "types_misc.h"
#include "errors.h"
#include "parser/plugins.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

static NxpElement *zero = NULL, *one = NULL, *mone = NULL;

NxpElement* nxp_element_zero(void) {
    return NXP_ELEMENT_REF(zero);
}

NxpElement* nxp_element_int_new_from_string(const gchar *string)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementInt));
    mpz_init_set_str(NXP_ELEMENT_INT(elem)->value, string, 10);
    return elem;
}

static inline NxpElement* nxp__element_int_new(nxp_int_t n)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementInt));
    mpz_init_set_si(NXP_ELEMENT_INT(elem)->value, n);
    return elem;
}

NxpElement* nxp_element_int_new(nxp_int_t n) 
{
    /* Cache for -1, 0, 1 */
    if (n == 0) 
	return NXP_ELEMENT_REF(zero);
    if (n == 1) 
	return NXP_ELEMENT_REF(one);
    if (n == -1) 
	return NXP_ELEMENT_REF(mone);

    return nxp__element_int_new(n);
}

/* Create a new NxpElementInt from a mpz_t number. The mpz_t number is not
 * copied, we keep it, because it can be very long to initialize a new mpz_t,
 * copy the number into it and then clear the old one. */
NxpElement* nxp_element_int_new_from_mpz(mpz_t i)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementInt));
    NXP_ELEMENT_INT(elem)->value[0] = i[0];
    return elem;
}

nxp_int_t nxp_element_int_get_ui(NxpElement *self, GError **error)
{
    NxpElementInt *value = NXP_ELEMENT_INT(self);

    if (mpz_fits_slong_p(value->value))
	return mpz_get_ui(value->value);
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_OVERFLOW, _("The integer you attempt to convert, is too big to fit a long integer"));
	return 0;
    }
}

nxp_int_t inline nxp_element_int_get_si(NxpElement *self, GError **error)
{
    NxpElementInt *value = NXP_ELEMENT_INT(self);

    if (mpz_fits_slong_p(value->value))
	return mpz_get_si(value->value);
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_OVERFLOW, _("The integer you attempt to convert, is too big to fit a long integer"));
	return 0;
    }
}

#define INT_INT(func, mpz_cmd)   static NxpElement* func(NxpOp op, NxpElement* a, NxpElement* b,\
  			                             NxpEvalContext *context, GError **error) {\
    mpz_t i; NxpElement *el;\
    mpz_init_set(i, NXP_ELEMENT_INT(a)->value); mpz_cmd(i, i, NXP_ELEMENT_INT(b)->value);\
    el = nxp_element_int_new_from_mpz(i);\
    return el; }

INT_INT(add, mpz_add);
INT_INT(sub, mpz_sub);
INT_INT(mul, mpz_mul);
INT_INT(rem, mpz_tdiv_r);
#undef INT_INT

#define NEW_INT_INT(func, mpz_cmd)   static NxpElement* func(NxpElementInt* a, NxpElementInt* b,\
  			                             NxpEvalContext *context, GError **error) {\
    mpz_t i; NxpElement *el;\
    mpz_init_set(i, NXP_ELEMENT_INT(a)->value); mpz_cmd(i, i, NXP_ELEMENT_INT(b)->value);\
    el = nxp_element_int_new_from_mpz(i);\
    return el; }

#undef NEW_INT_INT

NxpElement *nxp_element_int_mul(NxpOp op, NxpElement* a, NxpElement* b,
				NxpEvalContext *context, GError **error)
{
    return mul(op, a, b, context, error);
}

NxpElement *range(NxpOp op, NxpElement* a, NxpElement* b,
		  NxpEvalContext *context, GError **error)
{
    return nxp_element_vector_new_from_range(NXP_ELEMENT_INT(a), NXP_ELEMENT_INT(b), error);
}

static NxpElement* divide(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    mpq_t r;
    NxpElement *el;

    if (mpz_cmp_si(NXP_ELEMENT_INT(b)->value, 0)) {
	if (mpz_cmp_si(NXP_ELEMENT_INT(a)->value, 0)) {
	    mpq_init(r);
	    mpq_set_num(r, NXP_ELEMENT_INT(a)->value);
	    mpq_set_den(r, NXP_ELEMENT_INT(b)->value);
	    mpq_canonicalize(r);
	    if(mpz_cmp_si(mpq_denref(r), 1) == 0) {
		mpz_t z;
		mpz_init_set(z, mpq_numref(r));
		el = nxp_element_int_new_from_mpz(z);
		mpq_clear(r);
	    }
	    else
		el = nxp_element_frac_new_from_mpq(r);
	} else
	    return nxp_element_zero();
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
			_("divide by zero"));
	return NULL;
    }
    return el;
}

static NxpElement* eval_divide(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    NxpElement *el;
    mpq_t r;

    if (mpz_cmp_si(NXP_ELEMENT_INT(b)->value, 0)) {
	if (mpz_cmp_si(NXP_ELEMENT_INT(a)->value, 0)) {
	    mpq_init(r);
	    mpq_set_num(r, NXP_ELEMENT_INT(a)->value);
	    mpq_set_den(r, NXP_ELEMENT_INT(b)->value);
	    mpq_canonicalize(r);
	    if(mpz_cmp_si(mpq_denref(r), 1) == 0) {
		mpz_t z;
		mpz_init_set(z, mpq_numref(r));
		el = nxp_element_int_new_from_mpz(z);
	    }
	    else {
		mpfr_t d;
		mpfr_init_set_q(d, r, GMP_RNDN);
		el = nxp_element_real_new_from_mpfr(d);
	    }
	    mpq_clear(r);
	} else
	    return nxp_element_int_new(0);
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
			_("divide by zero"));
	return NULL;
    }
    return el;
}

static GString* str_func(NxpElement *el, int *precedence) {
    GString *gstr;
    char *str;

    gmp_asprintf(&str, "%Zd", NXP_ELEMENT_INT(el)->value);
    gstr = g_string_new(str);
    g_free(str);
    if (precedence)
	*precedence = 0;

    return gstr;
}

static NxpElement* nxp_int_and_or_int(NxpOp op, NxpElement *a, NxpElement *b,
				      NxpEvalContext *context, GError **error)
{
    NxpElement* retval;
    if (op == NXP_OP_SLASH_BSLASH) {
	if (mpz_cmp_ui(NXP_ELEMENT_INT(a)->value, 0))
	    retval = nxp_element_int_new(mpz_cmp_ui(NXP_ELEMENT_INT(b)->value, 0));
	else
	    retval = nxp_element_int_new(0);
    }
    else {
	if (mpz_cmp_ui(NXP_ELEMENT_INT(a)->value, 0))
	    retval = nxp_element_int_new(1);
	else
	    retval = nxp_element_int_new(mpz_cmp_ui(NXP_ELEMENT_INT(b)->value, 0));
    }
    return retval;
}

static NxpElement* eval_eq(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp(
			       NXP_ELEMENT_INT(a)->value,
			       NXP_ELEMENT_INT(b)->value) == 0);
}

static NxpElement* eval_neq(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp(
			       NXP_ELEMENT_INT(a)->value,
			       NXP_ELEMENT_INT(b)->value) != 0);
}

static NxpElement* eval_not(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp_si(
			       NXP_ELEMENT_INT(b)->value, 0));
}

static NxpElement* eval_gt(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp(
			       NXP_ELEMENT_INT(a)->value,
			       NXP_ELEMENT_INT(b)->value) > 0);
}

static NxpElement* eval_ge(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp(
			       NXP_ELEMENT_INT(a)->value,
			       NXP_ELEMENT_INT(b)->value) >= 0);
}

static NxpElement* eval_lt(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp(
			       NXP_ELEMENT_INT(a)->value,
			       NXP_ELEMENT_INT(b)->value) < 0);
}

static NxpElement* eval_le(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpz_cmp(
			       NXP_ELEMENT_INT(a)->value,
			       NXP_ELEMENT_INT(b)->value) <= 0);
}

static NxpElement* power(NxpOp op, NxpElement* a, NxpElement* b,
			 NxpEvalContext *context, GError **error) {
    mpz_t i;
    mpq_t q;
    GError *err = NULL;
    nxp_int_t pow = nxp_element_int_get_si(b, &err);
    NxpElement *el;
    int p = (pow >= 0)?pow:-pow;

    if (err) {
	g_clear_error(&err);
	g_set_error(error, nxp_error_domain, NXP_ERROR_OVERFLOW,
		    _("overflow in a power evaluation"));
	return NULL;
    }
    mpz_init(i);
    mpz_pow_ui(i, NXP_ELEMENT_INT(a)->value, p);
    if (pow < 0) {
	if (mpz_sgn(i) != 0) {
	    mpq_init(q);
	    mpq_set_z (q, i);
	    mpq_inv(q, q);
	    el = nxp_element_frac_new_from_mpq(q);
	}
	else {
	    el = NULL;
	    g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
		    _("divide by zero"));
	}
	mpz_clear(i);
    }
    else
	el = nxp_element_int_new_from_mpz(i);
    return el;
}

static NxpElement* eval_power(NxpOp op, NxpElement* a, NxpElement* b,
			      NxpEvalContext *context, GError **error) {
    mpz_t i;
    mpfr_t result;
    GError *err = NULL;
    nxp_int_t pow = nxp_element_int_get_si(b, &err);
    NxpElement *el;
    int p = (pow >= 0)?pow:-pow;

    if (err) {
	g_clear_error(&err);
	g_set_error(error, nxp_error_domain, NXP_ERROR_OVERFLOW,
		    _("overflow in a power evaluation"));
	return NULL;
    }
    mpz_init(i);
    mpz_pow_ui(i, NXP_ELEMENT_INT(a)->value, p);
    if (pow < 0) {
	if (mpz_sgn(i) != 0) {
	    mpfr_init_set_z(result, i, GMP_RNDN);
	    mpfr_ui_div(result, 1, result, GMP_RNDN);
	    el = nxp_element_real_new_from_mpfr(result);
	}
	else {
	    el = NULL;
	    g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
		    _("divide by zero"));
	}
	mpz_clear(i);
    }
    else
	el = nxp_element_int_new_from_mpz(i);
    return el;
}

static inline NxpElement* nxp_element_int_opposed(NxpElementNumber *self)
{
    NxpElement *el;
    mpz_t z;
    
    mpz_init_set(z, NXP_ELEMENT_INT(self)->value);
    mpz_neg(z, z);
    el = nxp_element_int_new_from_mpz(z);
    return el;
}

static NxpElement* uminus(NxpOp op, NxpElement* a, NxpElement* b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_int_opposed(NXP_ELEMENT_NUMBER(b));
}

static NxpElement* factorial(NxpOp op, NxpElement* a, NxpElement* b,
			     NxpEvalContext *context, GError **error) {
    mpz_t i;    NxpElement *el;

    if (mpz_cmp_si(NXP_ELEMENT_INT(a)->value, 0) < 0)
    {
  	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("Invalid argument to the factorial function (a non-negative integer must be provided)!"));
  	return NULL;
    }
    mpz_init(i);
    mpz_fac_ui(i, mpz_get_ui(NXP_ELEMENT_INT(a)->value));
    el = nxp_element_int_new_from_mpz(i);
    return el;
}

#define BIND(op,func)    nxp_bind_operator_function (op, 			\
  				NXP_ELEMENT_TYPE_INT, NXP_ELEMENT_TYPE_INT, \
  				func,  			\
  				func, 			\
  				0, 					\
				NULL); 

#define BIND_FUNC(op, eval, simp)    nxp_bind_operator_function (op, 	  \
  				NXP_ELEMENT_TYPE_INT, NXP_ELEMENT_TYPE_INT, \
  				eval,  		               	  \
  				simp,   	  	   	  \
  				0, 				  \
				NULL); 

#define BIND_RIGHT(op,func)    nxp_bind_operator_function (op, \
  				          NXP_ELEMENT_TYPE_INT, 0, func, \
  				          func, 0, NULL); 


#define BIND_LEFT(op,func)    nxp_bind_operator_function (op, \
  				0, NXP_ELEMENT_TYPE_INT, 	 \
  				func,        	 \
  				func,        	 \
  				0, NULL); 

static gboolean truth_func(NxpElement *self, NxpEvalContext *context, GError **error)
{
    return mpz_cmp_si(NXP_ELEMENT_INT(self)->value, 0);
}

static int nxp_element_int_compare_int(NxpElementNumber *self, nxp_int_t nb)
{
    return mpz_cmp_si(NXP_ELEMENT_INT(self)->value, nb);
}

static gboolean nxp_element_int_is_integer(NxpElementNumber *self)
{
    return TRUE;
}

static double nxp_element_int_get_value_as_double(NxpElementR *self)
{
    return mpz_get_d(NXP_ELEMENT_INT(self)->value);
}

static nxp_int_t nxp_element_int_get_value_as_int(NxpElementR *self)
{
    return mpz_get_si(NXP_ELEMENT_INT(self)->value);
}

NxpElement* nxp_element_int_sqrt_floor(NxpElement *self) {
    NxpElement *retval;
    mpz_t i;

    mpz_init(i);
    mpz_sqrt(i, NXP_ELEMENT_INT(self)->value);
    retval = nxp_element_int_new_from_mpz(i);
    return retval;
}

NxpElement* nxp_element_int_normalize(NxpElementInt *a,
				      mpz_t p,
				      NxpEvalContext *context,
				      GError **error)
{
    NxpElement *retval;
    mpz_t i, aa, p2, mp2;

    mpz_init(p2);
    mpz_cdiv_q_2exp(p2, p, 1);
    mpz_init(aa);
    mpz_mod(aa, a->value, p);
    mpz_init(i);
    mpz_init(mp2);
    mpz_neg(mp2, p2);

    if (mpz_cmp(aa, p2) > 0)
	mpz_sub(i, aa, p);
    else if (mpz_cmp(aa, mp2) <= 0)
	mpz_add(i, aa, p);
    else
	mpz_set(i, aa);
    retval = nxp_element_int_new_from_mpz(i);
    mpz_clear(aa);
    mpz_clear(p2);
    mpz_clear(mp2);
    return retval;
}


NxpElement* nxp_element_int_gcd(NxpElementInt *a, NxpElementInt *b,
				NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    mpz_t i;

    mpz_init(i);
    mpz_gcd(i, a->value, b->value);
    retval = nxp_element_int_new_from_mpz(i);
    return retval;
}

NxpElement* nxp_element_int_inverse(NxpElementInt *a, NxpElementInt *p,
				    NxpEvalContext *context, GError **error)
{
    NxpElement *retval;
    mpz_t i;

    mpz_init(i);
    mpz_invert(i, a->value, p->value);
    retval = nxp_element_int_new_from_mpz(i);
    return retval;
}

static void destroy(NxpElement *el) {
    mpz_clear(NXP_ELEMENT_INT(el)->value);
}

NxpElementType nxp_element_int_get_type(void)
{
    static int nxp_element_type_int = 0;

    if (nxp_element_type_int == 0) {
	nxp_element_type_int = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					nxp_element_type_int,
					sizeof(NxpElementRTypeInfo),
					NXP_ELEMENT_TYPE_R,
					sizeof(NxpElementRTypeInfo));
    }
    return nxp_element_type_int;
}

void nxp_element_int_type_init()
{
    NxpElementNumberTypeInfo *nb_typeinfo;
    NxpElementRTypeInfo *r_typeinfo;

    typeinfo = nxp_element_type_get_info(nxp_element_int_get_type());
    nb_typeinfo = (NxpElementNumberTypeInfo *)typeinfo;
    r_typeinfo = (NxpElementRTypeInfo *)typeinfo;

    typeinfo->name = g_strdup("int");
    typeinfo->eval_func = nxp_eval_return_self;
    typeinfo->simplify_func = nxp_eval_return_self;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = str_func;
    typeinfo->truth_func = truth_func;
    typeinfo->destroy_func = destroy;
    nb_typeinfo->compare_int = nxp_element_int_compare_int;
    nb_typeinfo->is_integer = nxp_element_int_is_integer;
    nb_typeinfo->opposed = nxp_element_int_opposed;
    r_typeinfo->get_value_as_double = nxp_element_int_get_value_as_double;
    r_typeinfo->get_value_as_int = nxp_element_int_get_value_as_int;

    nxp_parser_register_func(NXP_ATOM_CLASS_INT, nxp_element_int_new_from_string, 50);
    one = nxp__element_int_new(1);
    zero = nxp__element_int_new(0);
    mone = nxp__element_int_new(-1);

    BIND(NXP_OP_PLUS, add);
    BIND(NXP_OP_MINUS, sub);
    BIND(NXP_OP_STAR, mul);
    BIND_FUNC(NXP_OP_SLASH, eval_divide, divide);
    BIND_FUNC(NXP_OP_HAT, eval_power, power);
    BIND_LEFT(NXP_OP_MINUS_LEFT, uminus);
    BIND(NXP_OP_PERCENT, rem);


    BIND(NXP_OP_DOT3, range);
//    BIND_LEFT(NXP_OP_PLUS2_LEFT);
//    BIND_LEFT(NXP_OP_MINUS2_LEFT);
//    BIND_RIGHT(NXP_OP_PLUS2_RIGHT);
//    BIND_RIGHT(NXP_OP_MINUS2_RIGHT);

    BIND(NXP_OP_GT, eval_gt);
    BIND(NXP_OP_GT_EQ, eval_ge);
    BIND(NXP_OP_LT, eval_lt);
    BIND(NXP_OP_LT_EQ, eval_le);
    BIND(NXP_OP_EQ2, eval_eq);
    BIND(NXP_OP_EXCL_EQ, eval_neq);

    BIND(NXP_OP_SLASH_BSLASH, nxp_int_and_or_int);
    BIND(NXP_OP_BSLASH_SLASH, nxp_int_and_or_int);

    BIND_LEFT(NXP_OP_TILDE_LEFT, eval_not);
    BIND_RIGHT(NXP_OP_EXCL_RIGHT, factorial);
}
