/* -*- Mode: C; c-file-style: "stroustrup"; coding: latin-1-unix -*- */
#include <stdio.h>
#include "list.h"
#include "tree.h"
#include "int.h"
#include "message.h"

#include "errors.h"
#include "element_output.h"
#include "types_misc.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;



/**
 * nxp_element_list_new: -create list element
 * 
 * Create a new list element.
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_list_new(void)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementList));
    NXP_ELEMENT_LIST(elem)->list = NULL;
    NXP_ELEMENT_LIST(elem)->last = NULL;
    return elem;
}

/**
 * nxp_element_list_new_simple: -create list element
 * @el: A single element that constitutes the list.
 * 
 * Create a new list element.
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_list_new_simple(NxpElement *el)
{
    NxpElement *elem;

    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementList));
    NXP_ELEMENT_LIST(elem)->list = NULL;
    NXP_ELEMENT_LIST(elem)->last = NULL;
    nxp_element_list_append(elem, el);
    g_assert(NXP_ELEMENT_LIST(elem)->last->next == NULL);
    return elem;
}

/**
 * nxp_element_list_new_from_list: -create list element
 * @list: A GLib single linked list of NxpElements.
 *
 * @list becomes the new element property.
 *
 * Return value: the new element
 **/
NxpElement* nxp_element_list_new_from_list(GSList *list) {
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementList));
    NXP_ELEMENT_LIST(elem)->list = list;
    if (list) {
	NXP_ELEMENT_LIST(elem)->last = g_slist_last(list);
	g_assert(NXP_ELEMENT_LIST(elem)->last->next == NULL);
    }
    else
	NXP_ELEMENT_LIST(elem)->last = NULL;

    return elem;
}

static void nxp_destroy_list(NxpElement *elem)
{
    NxpElementList *lst = NXP_ELEMENT_LIST(elem);
    GSList *l;
    for (l = lst->list; l; l = l->next)
	NXP_ELEMENT_UNREF(NXP_ELEMENT(l->data));
    g_slist_free(lst->list);
}

NxpElement *nxp_element_list_insert(NxpElement *list, NxpElement *el, gint position)
{
    NxpElement *retval = nxp_element_list_new();
    GSList *l;
    int i;
    gboolean done = FALSE;

    if (position <= 0) {
	nxp_element_list_append(retval, el);
	nxp_element_list_append_list(retval, list);
	return retval;
    }
    for (i = 0, l = NXP_ELEMENT_LIST(list)->list; l; l = l->next, i++) {
	if (i == position) {
	    nxp_element_list_append(retval, el);
	    done = TRUE;
	    nxp_element_list_append(retval, l->data);
	}
	else
	    nxp_element_list_append(retval, l->data);
    }
    if (!done)
	nxp_element_list_append(retval, el);
    return retval;
}

void nxp_element_list_append(NxpElement *el_args, NxpElement *el)
{
    NxpElementList *list = NXP_ELEMENT_LIST(el_args);

    list->last = g_slist_append(list->last, NXP_ELEMENT_REF(el));
    if (!list->list)
	list->list = list->last;
    else
	list->last = list->last->next;
    g_assert(list->last->next == NULL);
}

void nxp_element_list_append_list(NxpElement *self, NxpElement *list1)
{
    GSList *l1 = NXP_ELEMENT_LIST(list1)->list;

    for ( ;l1; l1 = l1->next)
	nxp_element_list_append(self, l1->data);
}

void nxp_element_list_prepend(NxpElement *el_args, NxpElement *el)
{
    NxpElementList *list = NXP_ELEMENT_LIST(el_args);

    list->list = g_slist_prepend(list->list, NXP_ELEMENT_REF(el));
    if (!list->last)
	list->last = list->list;
}

int nxp_element_list_contains(NxpElement *list, NxpElement *el,
			NxpEvalContext *context, GError **error)
{
    GSList *l = NXP_ELEMENT_LIST(list)->list;
    int index = 0;

    g_return_val_if_fail(NXP_ELEMENT(el), 0);

    while (l) {
	if (nxp_element_eq(el, NXP_ELEMENT(l->data), context, error))
	    return index;
	else {
	    index++;
	    l = l->next;
	}
    }
    return -1;
}

/**
 * nxp_element_list_extract:
 * @list: The list from which we extract a new list.
 * @index1: index of the first element to put in the new list.
 * @index2: index of the last element to put in the new list. If it is smaller than @index1, all the elements after the first one are put in the new list.
 *
 * Return value: The new list.
 **/
NxpElement *nxp_element_list_extract(NxpElement *list, int index1, int index2)
{
    int i;
    NxpElement *retval;
    GSList *lst = NXP_ELEMENT_LIST(list)->list;

    retval = nxp_element_list_new();
    for (i = 0; i < index1; i++)
	lst = lst->next;
    if (index2 >= index1)
	for ( ; lst && i <= index2; i++, lst = lst->next)
	    nxp_element_list_append(retval, lst->data);
    else
	for ( ; lst; lst = lst->next)
	    nxp_element_list_append(retval, lst->data);

    return retval;
}

gboolean nxp_element_list_same_content(NxpElement *list1, NxpElement *list2, NxpEvalContext *context, GError **error)
{
    GSList *lst1, *lst2, *l1, *l2;
    gboolean *tab;
    int l, i;

    lst1 = NXP_ELEMENT_LIST(list1)->list;
    lst2 = NXP_ELEMENT_LIST(list2)->list;

    l =  g_slist_length(lst2);
    tab = g_new0(gboolean, l);

    for (l1 = lst1; l1; l1 = l1->next) {
	for (i = 0, l2 = lst2; l2; i++, l2 = l2->next) {
	    if (!tab[i]) {
		if (nxp_element_eq(l1->data, l2->data, context, error)) {
		    tab[i] = TRUE;
		    break;
		}
	    }
	}
    }
    for (i = 0; i < l; i++)
	if (!tab[i]) {
	    g_free(tab);
	    return FALSE;
    }
    g_free(tab);
    return TRUE;
}
		

static GString* str_or_repr_func(NxpElement *elem, int *precedence, int str_or_repr)
{
    GString *string = g_string_new("[");
    NxpElement *el;
    NxpElementList* lst;
    GSList *l;
    GString* old;

    if (precedence) *precedence = 0;

    lst = NXP_ELEMENT_LIST(elem);
    for (l = lst->list; l ; l = l->next) {
	el = NXP_ELEMENT(l->data);
	if (str_or_repr == 1)
	    old = nxp_element_str(el);
	else
	    old = nxp_element_repr(el);
	string = g_string_append(string, old->str);
	if (l->next)
	    string = g_string_append(string, ", ");
	g_string_free(old, TRUE);
    }
    g_string_append(string, "]");
    return string;
}

static GString *str_func(NxpElement *elem, int *prec)
{
    return str_or_repr_func(elem, prec, 1);
}


static GString *repr_func(NxpElement *elem, int *prec)
{
    return str_or_repr_func(elem, prec, 2);
}

#define BIND(op)    nxp_bind_operator_function(op,	\
					       NXP_ELEMENT_TYPE_LIST, \
					       NXP_ELEMENT_TYPE_LIST, \
    			                       map_op_eval, map_op_simplify,	\
					       0, NULL);

#define BIND_FUNC(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_LIST, \
							 NXP_ELEMENT_TYPE_LIST, \
    			                                 eval, symp, 0, NULL);
#define BIND_LEFT(eval,symp,op)    nxp_bind_operator_function(op, 0,         \
                                                         NXP_ELEMENT_TYPE_LIST, \
   			                                 eval, symp, 0, NULL);

NxpElement* nxp_eval_list(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElement *retval, *el;
    GSList *lst;
    GError *err = NULL;
    gboolean err_test = FALSE;

    retval = nxp_element_list_new();
    for (lst = NXP_ELEMENT_LIST(elem)->list; lst; lst = lst->next) {
	el = nxp_element_eval(NXP_ELEMENT(lst->data), context, &err);
	if (err) {
	    if (!err_test) {
		err_test = TRUE;
		g_propagate_error(error, err);
	    }
	    else {
		g_clear_error(&err);
	    }
	    err = NULL;
	    el = NXP_ELEMENT_REF(lst->data);
	}
	g_assert(el != NULL);
	nxp_element_list_append(retval, el);
	NXP_ELEMENT_UNREF(el);
    }
    return retval;
}

NxpElement* nxp_simplify_list(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    NxpElement *retval, *el;
    GSList *lst = NXP_ELEMENT_LIST(elem)->list;
    GError *err = NULL;

    retval = nxp_element_list_new();
    for (lst = NXP_ELEMENT_LIST(elem)->list; lst; lst = lst->next) {
	el = nxp_element_simplify(NXP_ELEMENT(lst->data), context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(retval);
	    return NULL;
	}
	g_assert(el);
	nxp_element_list_append(retval, el);
	NXP_ELEMENT_UNREF(el);
    }
    return retval;
}

static void
iterator_func(NxpElement *elem,
              NxpElementVisitorFunc func,
              gpointer user_data,
              NxpEvalContext *context)
{
    GSList *l = NXP_ELEMENT_LIST(elem)->list;

    for (; l; l = l->next) {
        NxpElement *item = NXP_ELEMENT(l->data);
        func(item, user_data);
        if (item->type_info->iterator_func)
            item->type_info->iterator_func(item, func, user_data, context);
    }
}

/*
static NxpElement* map_op_simplify(NxpOp op,
			  NxpElement *a, NxpElement *b,
			  NxpEvalContext *context, GError **error) {
    NxpElementVector *result;
    gint i;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(b), NULL);

    if (NXP_ELEMENT_VECTOR(a)->size != NXP_ELEMENT_VECTOR(b)->size) {
	// TODO -- GERROR SIZES DIFERENT
    } else {
	result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(NXP_ELEMENT_VECTOR(a)->size,
							   NULL));
	for (i = 0; i < NXP_ELEMENT_VECTOR(a)->size; i++)
	{
	    _element_vector_index(result, i) = nxp_element_tree_simplify_op(
			    op,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	}
	return NXP_ELEMENT(result);
    }
    return NULL;
}

static NxpElement* map_op_eval(NxpOp op,
			       NxpElement *a, NxpElement *b,
			       NxpEvalContext *context, GError **error) {
    NxpElementVector *result;
    gint i;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(b), NULL);

    if (NXP_ELEMENT_VECTOR(a)->size != NXP_ELEMENT_VECTOR(b)->size) {
	g_warning ("Different sizes !!!\n");
	// TODO -- GERROR SIZES DIFERENT
    } else {
	result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(NXP_ELEMENT_VECTOR(a)->size,
							   NULL));
	for (i = 0; i < NXP_ELEMENT_VECTOR(a)->size; i++)
	{
	    _element_vector_index(result, i) = nxp_element_tree_eval_op(
			    op,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	}
	return NXP_ELEMENT(result);
    }
    return NULL;
}
*/

static gboolean truth_func(NxpElement *lst, NxpEvalContext *context, GError **error)
{
    return NXP_ELEMENT_LIST(lst)->list && NXP_ELEMENT_LIST(lst)->list->data;
}

static gboolean dependson_func(NxpElement *self, NxpElement *variable, NxpEvalContext *context, GError **error)
{
    GSList *l;
    GError *err = NULL;
    gboolean retval = FALSE;

    l = NXP_ELEMENT_LIST(self)->list;
    while (l && !retval) {
	retval = nxp_element_depends_on(NXP_ELEMENT(l->data), variable, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    break;
	}
	l = l->next;
    }
    return retval;
}

static NxpElement* substitute_func(NxpElement *self, NxpElement *var, NxpElement *el, NxpEvalContext *context, GError **error)
{
    NxpElement *retval, *tmp;
    GSList *l;
    GError *err = NULL;

    retval = nxp_element_list_new();
    for (l = NXP_ELEMENT_LIST(self)->list; l; l = l->next) {
	tmp = nxp_element_substitute(NXP_ELEMENT(l->data), var, el, context, &err);
	if (err) {
	    NXP_ELEMENT_UNREF(retval);
	    g_propagate_error(error, err);
	    return NULL;
	}
	nxp_element_list_append(retval, tmp);
	NXP_ELEMENT_UNREF(tmp);
    }
    return retval;
}

static NxpElement* simp_eq2(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error) {
    GSList *la, *lb;

    la = NXP_ELEMENT_LIST(a)->list;
    lb = NXP_ELEMENT_LIST(b)->list;
    while (la && lb && nxp_element_eq(la->data, lb->data, context, error)) {
	la = la->next;
	lb = lb->next;
    }
    if (!la && !lb)
	return nxp_element_int_new(1);
    return nxp_element_int_new(0);
}

int inline nxp_element_list_length(NxpElement *self)
{
    return g_slist_length(NXP_ELEMENT_LIST(self)->list);
}

int nxp_element_list_get_type(void)
{
    static int nxp_element_type_list = 0;

    if (nxp_element_type_list == 0) {
	nxp_element_type_list = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_list,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_list;
}

void nxp_element_list_type_init() 
{
    typeinfo = nxp_element_type_get_info(nxp_element_list_get_type());
    typeinfo->name = g_strdup("list");
    typeinfo->destroy_func = nxp_destroy_list;
    typeinfo->eval_func = nxp_eval_list;
    typeinfo->simplify_func = nxp_simplify_list;
    typeinfo->derive_func = NULL;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = repr_func;
    typeinfo->truth_func = truth_func;
    typeinfo->iterator_func = iterator_func;
    typeinfo->dependson_func = dependson_func;
    typeinfo->substitute_func = substitute_func;
}

void nxp_element_list_operator_init(void)
{
    BIND_FUNC(NULL, simp_eq2, NXP_OP_EQ2);

//    BIND(NXP_OP_PLUS);
//    BIND(NXP_OP_MINUS);
//    BIND(NXP_OP_STAR);
//    BIND(NXP_OP_SLASH);
//    BIND_LEFT(minus, NULL, NXP_OP_MINUS_LEFT);
    //    NXP_OP_HAT

    /// not implementable ///////
    //
    //    NXP_OP_DOT3
    //    NXP_OP_PLUS2_LEFT
    //    NXP_OP_MINUS2_LEFT
    //    NXP_OP_PLUS2_RIGHT
    //    NXP_OP_MINUS2_RIGHT
    /////////////////////////////

//    BIND_FUNC( gt, gt, NXP_OP_GT);
//    BIND_FUNC( le, le, NXP_OP_LT_EQ);
//    BIND_FUNC( lt, lt, NXP_OP_LT);
//    BIND_FUNC( ge, ge, NXP_OP_GT_EQ);

    // == and != compares ALL
//    BIND_FUNC( eq, eq, NXP_OP_EQ2);
//    BIND_FUNC( ne, NULL, NXP_OP_EXCL_EQ);

    // These two compares sizes, only
//    BIND_FUNC( and, and, NXP_OP_SLASH_BSLASH);
//    BIND_FUNC( or,  or,  NXP_OP_BSLASH_SLASH);
//    BIND_LEFT( not, not, NXP_OP_TILDE_LEFT);

    ///// not implementable ///
    //
    //    NXP_OP_EXCL_RIGHT
    //    NXP_OP_EQ
    //    NXP_OP_COLON_EQ
    ///////////////////////////
}
