/* -*- Mode: C; c-file-style: "stroustrup"; coding: latin-1-unix -*- */
#include <stdio.h>
#include "vector.h"

#include "message.h"

#include "errors.h"
#include "number.h"
#include "complex.h"
#include "matrix.h"
#include "tree.h"
#include "element_output.h"
#include "types_misc.h"
#include "void.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

/**
 * nxp_element_vector_new: -create vector element
 * @size: the number of elements in the vector
 * @data: an array of pointers to each element of the vector; this block of
 *        memory should have a size of at least (and, if possible, no more than) 
 *        @size*sizeof(&NxpElement*); if %NULL, a new array of pointers (initialized to %NULL)
 *        will be created to hold @size elements; Note: if @data is non null, the vector element
 *        will take ownership (keep) of this array.
 * 
 * Create a new vector element.
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_vector_new(guint size, NxpElement **data)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementVector));
    NXP_ELEMENT_VECTOR(elem)->size = size;
    if (data)
	NXP_ELEMENT_VECTOR(elem)->data = data;
    else
	NXP_ELEMENT_VECTOR(elem)->data = size > 0? g_new0(NxpElement *, size) : NULL;
    return elem;
}

/**
 * nxp_element_vector_new_from_range: -create vector element
 * @low: the first element in the vector
 * @high: the last element in the vector
 *
 * @low and @high are two integers such that @low < @high. This function
 * creates the vector {@low, @low + 1, @low + 2, ..., @high}.
 *
 * Return value: the new element
 **/
NxpElement* nxp_element_vector_new_from_range(NxpElementInt *low, NxpElementInt *high, GError **error) {
    mpz_t value, val;

    if (mpz_cmp(low->value, high->value) <= 0) {
	NxpElement *vec;
	nxp_int_t i, size;
	mpz_init_set(value, high->value);
	mpz_sub(value, value, low->value);
	mpz_add_ui(value, value, 1);

	if (mpz_fits_slong_p(value))
	    size = mpz_get_si(value);
	else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_OVERFLOW, _("The vector size requested is too big"));
	    mpz_clear(value);
	    return NULL;
	}

	vec = nxp_element_vector_new(size, NULL);
	mpz_set(value, low->value);
	for (i = 0; i < size; i++, mpz_add_ui(value, value, 1)) {
	    mpz_init_set(val, value);
	    NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(vec), i) = nxp_element_int_new_from_mpz(val);
	}
	mpz_clear(value);
	return vec;
    } else {
	return nxp_element_vector_new(0, NULL);
    }
}

/* nxp_element_vector_extract_range:
 * @v: element
 * @from: start position
 * @to: end position
 *
 * Extract the sub-vector from the vector v from the index @from to the index
 * @to.
 **/
NxpElement* nxp_element_vector_extract_range(NxpElement *v, int from, int to)
{
    NxpElement *retval;
    int i;
    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(v), NULL);
    g_return_val_if_fail(from <= to, NULL);
    g_return_val_if_fail(from >= 0 && to < NXP_ELEMENT_VECTOR(v)->size, NULL);

    retval = nxp_element_vector_new(to - from + 1, NULL);
    for (i = from; i <= to; i++)
	_element_vector_index(NXP_ELEMENT_VECTOR(retval), i - from) =
				_element_vector_get(NXP_ELEMENT_VECTOR(v), i);
				
    return retval;
}

/* nxp_element_vector_made_of:
 * @el: a vector
 * @type: a type
 *
 * Test if all the coordinates of the vector are of the type @type.
 *
 * This function is useful when we debug...
 *
 * Returned value: TRUE if the test succeeds, FALSE if the test fails.
**/
gboolean nxp_element_vector_made_of(NxpElement *el, NxpElementType type)
{
    int i;
    NxpElementVector *v = (NxpElementVector *)el;
    GString *string;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(el), FALSE);
    for (i = v->size - 1; i >= 0; i--)
	if (!nxp_element_check_type(_element_vector_index(v, i), type)) {
	    string = nxp_element_repr(el);
	    fprintf(stderr, "Vector is : %s", string->str);
	    g_string_free(string, TRUE);
	    g_print("Type seen :\n");
	    nxp_elem_printer(_element_vector_index(v, i));
	    return FALSE;
	}
    return TRUE;
}

static void nxp_destroy_vector(NxpElement *elem)
{
    NxpElementVector *vec = NXP_ELEMENT_VECTOR(elem);
    int i;
    for (i = 0; i < vec->size; ++i)
	NXP_ELEMENT_XUNREF(vec->data[i]);
    g_free(vec->data);
}

/**
 * _element_vector_set:
 * @vec: vector
 * @index: position
 * @element: value of element (will keep a reference)
 * 
 * Set the value of a specific element of the vector. Warning: vectors
 * are immutable -- they should not be changed after being created
 * (and initialized) !
 **/
void _element_vector_set(NxpElementVector *vec, guint index, NxpElement *element) 
{
    /* make sure we have a vector */
    g_return_if_fail(vec);
    /* make sure we have an element */
    g_return_if_fail(element);
    /* make sure we are not exceeding dimmensions */
    g_return_if_fail(index < vec->size);
    /* make sure the vector doesn't already have an element there */
    g_return_if_fail(!_element_vector_index(vec, index));
    _element_vector_index(vec, index) = NXP_ELEMENT_REF(element);
}



/**
 * nxp_element_vector_get:
 * @vec: the vector element
 * @index: index number (starts at 0)
 * 
 * Get a specific element of the vector.
 *
 * Return value: a new reference to the element at the specified
 * position of the vector if it has one, or the void element if there
 * is no element in that position.
 **/
NxpElement* _element_vector_get(NxpElementVector *vec, guint index) 
{
    NxpElement *el;
    /* make sure we have a vector */
    g_return_val_if_fail(vec, NULL);
    /* make sure we are not exceeding dimmensions */
    g_return_val_if_fail(index < vec->size, NULL);
    el = _element_vector_index(vec, index);
    return el? NXP_ELEMENT_REF(el) : nxp_element_void_new();
}



/**
 * nxp_element_vector_insert:
 * @vector: vector 
 * @index: position (starts at 0)
 * @element: element
 * 
 * Insert element in vector at position @index. A new vector is
 * created with changes from the original one.
 * 
 * Return value: a new vector
 **/
NxpElement* nxp_element_vector_insert(NxpElement* vector,
				      guint index, NxpElement* element)
{
    NxpElementVector *v;
    int i, size;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(vector), NULL);
    g_return_val_if_fail(NXP_ELEMENT(element), NULL);
    g_return_val_if_fail(index <= NXP_ELEMENT_VECTOR(vector)->size, NULL);

    size  = NXP_ELEMENT_VECTOR(vector)->size + 1;

    v = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (i = 0; i < index; i++)
	_element_vector_set(v, i,
			_element_vector_index(NXP_ELEMENT_VECTOR(vector), i));
    _element_vector_set(v, index, element);
    for (i = index + 1; i < size; i++)
	_element_vector_set(v, i,
		    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i - 1));
    return NXP_ELEMENT(v);
}

/**
 * nxp_element_vector_real_put:
 * @vector: vector 
 * @element: element
 * @index: position (starts at 0)
 * 
 * Put element in vector, expanding it if necessary. A new vector is
 * created with changes from the original one.
 * 
 * Return value: a new vector
 **/
NxpElement* nxp_element_vector_real_put(NxpElementVector* vector,
					NxpElement* element, guint index)
{
    NxpElementVector *v;
    NxpElement *el;
    int i, size;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(NXP_ELEMENT(vector)), NULL);
    g_return_val_if_fail(NXP_ELEMENT(element), NULL);
    size  = MAX(vector->size, index + 1);

    v = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (i = 0; i < size; i++)
    {
	if (i == index)
	    el = NXP_ELEMENT_REF(element);
	else {
	    if (i < vector->size) {
		el = NXP_ELEMENT_REF(_element_vector_index(vector, i));
	    } else {
		el = NULL;
	    }
	}
	_element_vector_index(v, i) = el;
    }
    return NXP_ELEMENT(v);
}

/**
 * nxp_element_vector_set_size:
 * @v: vector
 * @size: positive integer
 *
 * Modify the size of the vector @v and keep the first @size elements of the
 * vector. If @v is too short, it is completed with zeros.
 *
 * Return value: a new vector
 **/
NxpElement* nxp_element_vector_set_size	(NxpElement *v, guint size)
{
    NxpElement *vector;
    int s = (NXP_ELEMENT_VECTOR(v)->size < size) ? NXP_ELEMENT_VECTOR(v)->size : size;
    int i;

    vector = nxp_element_vector_new(size, NULL);
    for (i = 0; i < s; i++)
	_element_vector_set(NXP_ELEMENT_VECTOR(vector), i,
			    _element_vector_index(NXP_ELEMENT_VECTOR(v), i));
    if (i < size - 1)
	for ( ; i < size; i++)
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), i) = nxp_element_int_new(0);
    return vector;
}

/**
 * vector_times_matrix:
 * @vec: 
 * @mat: 
 * 
 * Multiply vector @vec (on the left) by matrix @mat (on the
 * right). The vector is assumed to be horizontal or vertical,
 * whichever allows multiplication by algebra rules -- in this case,
 * it has to be horizontal.
 * 
 * Return value: a matrix with the result
 **/
NxpElement *vector_times_matrix(NxpElementVector *vec, NxpElementMatrix *mat, NxpEvalContext *context, GError **error)
{
    NxpElementMatrix *result;
    NxpElement *tree, *el;
    gint i, j;

    g_return_val_if_fail(NXP_ELEMENT_VECTOR(vec), NULL);
    g_return_val_if_fail(NXP_ELEMENT_MATRIX(mat), NULL);

    if (vec->size != mat->height) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("Dimensions don't agree: vector size %i := matrix height (%i)"), vec->size, mat->height);
	return NULL;
    }

    /* allocate matrix for result height = 1; width = mat->width*/
    result = NXP_ELEMENT_MATRIX(nxp_element_matrix_new(mat->width, 1, NULL));
    for (j = 0; j < mat->width; j++) {
	tree = NULL;
	for (i = 0; i < vec->size; i++) {
	    el = elem_TIMES(_element_vector_index(vec, i),
			    _element_matrix_index(mat, i, j));
	    if (!tree)
		tree = el;
	    else 
		tree = elem_PLUS(tree, el);
	}
	el = nxp_element_simplify(tree, context, NULL);
	NXP_ELEMENT_UNREF(tree);
	_element_matrix_set(result, 0, j, el);
	NXP_ELEMENT_UNREF(el);
    }
    return NXP_ELEMENT(result);
}

static GString* repr_func(NxpElement *elem, int *precedence)
{
    GString *string = g_string_new("{");
    int i;
    NxpElement *el;
    NxpElementVector* v;
    GString* old;

    if (precedence) *precedence = 0;

    v = NXP_ELEMENT_VECTOR(elem);
    for (i = 0; i < v->size; i++) {
	el = _element_vector_get(v, i);
	old = nxp_element_str(el); 
	NXP_ELEMENT_UNREF(el);
	string = g_string_append(string, old->str);
	string = g_string_append(string, ", ");
	g_string_free(old, TRUE);
    }
    if (v->size) string = g_string_truncate(string, (string)->len - 2);
    string = g_string_append_c(string, '}');
    
    return string;
}


static GString* str_func(NxpElement *elem, int *precedence)
{
    GString *string = g_string_new("{");
    int i;
    NxpElement *el;
    NxpElementVector* v;
    GString* old;

    if (precedence) *precedence = 0;

    v = NXP_ELEMENT_VECTOR(elem);
#define MAX_PRINTED_VECTOR_SIZE 30
    for (i = 0; i < v->size; i++) {
	el = _element_vector_get(v, i);
	old = nxp_element_str(el); 
	NXP_ELEMENT_UNREF(el);
	string = g_string_append(string, old->str);
	string = g_string_append(string, ", ");
	g_string_free(old, TRUE);
	if (i > MAX_PRINTED_VECTOR_SIZE) {
	    if (v->size - 10 > i) {
		string = g_string_append(string, "(...), ");
		i = v->size - 10;
	    }
	}
    }
    if (v->size) string = g_string_truncate(string, (string)->len - 2);
    string = g_string_append_c(string, '}');

    return string;
}

#define BIND(op)    nxp_bind_operator_function(op,	\
					       NXP_ELEMENT_TYPE_VECTOR, \
					       NXP_ELEMENT_TYPE_VECTOR, \
    			                       map_op_eval, map_op_simplify,	\
					       0, NULL);

#define BIND_FUNC(eval,symp,op)    nxp_bind_operator_function(op,                 \
							 NXP_ELEMENT_TYPE_VECTOR, \
							 NXP_ELEMENT_TYPE_VECTOR, \
    			                                 eval, symp, 0, NULL);
#define BIND_LEFT(eval,symp,op)    nxp_bind_operator_function(op, 0,         \
                                                         NXP_ELEMENT_TYPE_VECTOR, \
   			                                 eval, symp, 0, NULL);

static NxpElement* minus(NxpOp op,
			 NxpElement *a, NxpElement *b,
			 NxpEvalContext *context,
			 GError **error) {
    NxpElementVector *result;
    NxpElement *el, *tree;
    gint i;

    result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(NXP_ELEMENT_VECTOR(b)->size,
						       NULL));
    for (i = 0; i < NXP_ELEMENT_VECTOR(b)->size; i++)
    {
	tree = elem_UMINUS(_element_vector_index(NXP_ELEMENT_VECTOR(b), i));
	el = nxp_element_eval(tree, context, NULL);
	NXP_ELEMENT_UNREF(tree);
	_element_vector_index(result, i) = el;
    }
    return NXP_ELEMENT(result);

}

static NxpElement* minus_simp(NxpOp op,
			 NxpElement *a, NxpElement *b,
			 NxpEvalContext *context,
			 GError **error) {
    NxpElementVector *result;
    gint i;
    GError *err = NULL;

    result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(NXP_ELEMENT_VECTOR(b)->size,
						       NULL));
    for (i = 0; i < NXP_ELEMENT_VECTOR(b)->size; i++) {
	_element_vector_index(result, i) = nxp_element_tree_simplify_op(op,
			NULL, _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			context, &err);
	if (err) {
	    NXP_ELEMENT_UNREF(result);
	    g_propagate_error(error, err);
	    break;
	}
    }
    return NXP_ELEMENT(result);
}

static NxpElement* not(NxpOp op, NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error) {
    return nxp_element_int_new(!NXP_ELEMENT_VECTOR(b)->size);
}

/* Interesting function when we compares powers of mpoly elements. It goes much
 * faster than nxp_element_vector_compare */
gint nxp_element_vector_int_compare(NxpElement *a, NxpElement *b)
{
    NxpElementVector* aa = NXP_ELEMENT_VECTOR(a);
    NxpElementVector* bb = NXP_ELEMENT_VECTOR(b);
    guint s = aa->size < bb->size ? aa->size : bb->size;
    int i;
    nxp_int_t nba, nbb;

    for (i = 0; i < s; i++) {
	nba = nxp_element_int_get_si(_element_vector_index(aa, i), NULL);
	nbb = nxp_element_int_get_si(_element_vector_index(bb, i), NULL);
	if (nba > nbb)
	    return 1;
	else if (nba < nbb)
	    return -1;
    }
    return 0;
}

gint nxp_element_vector_compare(NxpElement *a, NxpElement *b)
{
    NxpElementVector* aa = NXP_ELEMENT_VECTOR(a);
    NxpElementVector* bb = NXP_ELEMENT_VECTOR(b);
    guint s = aa->size < bb->size ? aa->size : bb->size;
    int i;
    NxpElement *el_gt;

    for (i = 0; i < s; i++) {
	if (!nxp_element_eq(_element_vector_index(aa, i),
			    _element_vector_index(bb, i),
			    NULL, NULL)) {
	    el_gt = nxp_element_tree_simplify_op(NXP_OP_GT,
				_element_vector_index(NXP_ELEMENT_VECTOR(a), i),
				_element_vector_index(NXP_ELEMENT_VECTOR(b), i),
				NULL, NULL);
	    if (nxp_element_is_true(el_gt, NULL, NULL)) {
		NXP_ELEMENT_UNREF(el_gt);
		return 1;
	    }
	    else {
		NXP_ELEMENT_UNREF(el_gt);
		return -1;
	    }
	}
    }
    return 0;
}

static NxpElement* eq(NxpOp op,
		      NxpElement *a, NxpElement *b,
		      NxpEvalContext *context,
		      GError **error) {
    NxpElementVector* aa = NXP_ELEMENT_VECTOR(a);
    NxpElementVector* bb = NXP_ELEMENT_VECTOR(b);
    if (aa->size == bb->size) {
	int i;
	for (i = 0; i < aa->size; i++) {
	    if (!nxp_element_eq(_element_vector_index(aa, i),
				_element_vector_index(bb, i),
				context, error))
		return nxp_element_int_new(0);
	}
    }
    else
	return nxp_element_int_new(0);
    return nxp_element_int_new(1);
}

static NxpElement* ne(NxpOp op,
		      NxpElement *a, NxpElement *b,
		      NxpEvalContext *context,
		      GError **error) {
    NxpElement* r = eq(op, a, b, context, error);
    int res;
    if (NXP_ELEMENT_INT(r)->value) {
	res = 0;
    } else {
	res = 1;
    }
    NXP_ELEMENT_UNREF(r);
    return nxp_element_int_new(res);
}

static NxpElement* gt(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error) {
    guint s = NXP_ELEMENT_VECTOR(a)->size < NXP_ELEMENT_VECTOR(b)->size ?
		    NXP_ELEMENT_VECTOR(a)->size : NXP_ELEMENT_VECTOR(b)->size;
    int i;
    NxpElement *el_gt;

    for (i = 0; i < s; i++) {
	el_gt = nxp_element_tree_simplify_op(NXP_OP_GT,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	if (!NXP_ELEMENT_IS_NUMBER(el_gt) || nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(el_gt), 0) == 0)
	    return el_gt;
	else
	    NXP_ELEMENT_UNREF(el_gt);
    }
    return nxp_element_int_new(1);
}

static NxpElement* lt(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error) {
    guint s = NXP_ELEMENT_VECTOR(a)->size < NXP_ELEMENT_VECTOR(b)->size ?
		    NXP_ELEMENT_VECTOR(a)->size : NXP_ELEMENT_VECTOR(b)->size;
    int i;
    NxpElement *el_lt;

    for (i = 0; i < s; i++) {
	el_lt = nxp_element_tree_simplify_op(NXP_OP_LT,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	if (!NXP_ELEMENT_IS_NUMBER(el_lt) ||
	    nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(el_lt), 0) == 0)
	    return el_lt;
	else
	    NXP_ELEMENT_UNREF(el_lt);
    }
    return nxp_element_int_new(1);
}

static NxpElement* ge(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error) {
    guint s = NXP_ELEMENT_VECTOR(a)->size < NXP_ELEMENT_VECTOR(b)->size ?
		    NXP_ELEMENT_VECTOR(a)->size : NXP_ELEMENT_VECTOR(b)->size;
    int i;
    NxpElement *el_ge;

    for (i = 0; i < s; i++) {
	el_ge = nxp_element_tree_simplify_op(NXP_OP_GT_EQ,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	if (!NXP_ELEMENT_IS_NUMBER(el_ge) || nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(el_ge), 0) == 0)
	    return el_ge;
	else
	    NXP_ELEMENT_UNREF(el_ge);
    }
    return nxp_element_int_new(1);
}

static NxpElement* le(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error) {
    guint s = NXP_ELEMENT_VECTOR(a)->size < NXP_ELEMENT_VECTOR(b)->size ?
		    NXP_ELEMENT_VECTOR(a)->size : NXP_ELEMENT_VECTOR(b)->size;
    int i;
    NxpElement *el_le;

    for (i = 0; i < s; i++) {
	el_le = nxp_element_tree_simplify_op(NXP_OP_LT_EQ,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	if (!NXP_ELEMENT_IS_NUMBER(el_le) ||
	    nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(el_le), 0) == 0)
	    return el_le;
	else
	    NXP_ELEMENT_UNREF(el_le);
    }
    return nxp_element_int_new(1);
}

static NxpElement* and(NxpOp op, NxpElement *a, NxpElement *b, 
		       NxpEvalContext *context, GError **error) {
    int siza,sizb;
    siza = NXP_ELEMENT_VECTOR(a)->size;
    sizb = NXP_ELEMENT_VECTOR(b)->size;
    return nxp_element_int_new(siza?sizb:siza);
}
static NxpElement* or(NxpOp op, NxpElement *a, NxpElement *b, 
		      NxpEvalContext *context, GError **error) {
    int siza,sizb;
    siza = NXP_ELEMENT_VECTOR(a)->size;
    sizb = NXP_ELEMENT_VECTOR(b)->size;
    return nxp_element_int_new(siza?siza:sizb);
}



NxpElement* nxp_eval_vector(NxpElement *elem, NxpEvalContext *context, GError **error) {
    GError *err = NULL;
    NxpElement *el;

    NxpElementVector* vec = NXP_ELEMENT_VECTOR(elem);
    int size = NXP_ELEMENT_VECTOR_SIZE(vec);
    NxpElement* new = nxp_element_vector_new(size, NULL);
    int i;
    for (i=0; i<size; i++) {
	el = nxp_element_eval(NXP_ELEMENT_VECTOR_ELEM(vec, i), context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(new);
	    return NULL;
	}
	_element_vector_index(NXP_ELEMENT_VECTOR(new),i) = el;
    }
    return new;

}

NxpElement* nxp_simplify_vector(NxpElement *elem, NxpEvalContext *context, GError **error) {
	NxpElementVector* vec = NXP_ELEMENT_VECTOR(elem);
	int size = NXP_ELEMENT_VECTOR_SIZE(vec);
	NxpElement* new = nxp_element_vector_new(size, NULL);
	int i;
	for (i=0; i<size; i++) {
	    _element_vector_index(NXP_ELEMENT_VECTOR(new),i) = 
		nxp_element_simplify(NXP_ELEMENT_VECTOR_ELEM(vec, i), context, NULL); /* FIXME: use error */
	}
	return new;

}

static void
iterator_func(NxpElement *elem,
              NxpElementVisitorFunc func,
              gpointer user_data,
              NxpEvalContext *context)
{
    NxpElementVector *vec = NXP_ELEMENT_VECTOR(elem);
    int i;
  
    for (i = 0; i < vec->size; ++i)
    {
        NxpElement *item = NXP_ELEMENT(vec->data[i]);
        func(item, user_data);
        if (item->type_info->iterator_func)
            item->type_info->iterator_func(item, func, user_data, context);
    }
}

static gboolean dependson_func(NxpElement *self, NxpElement *variable, NxpEvalContext *context, GError **error)
{
    GError *err = NULL;
    NxpElementVector *v = NXP_ELEMENT_VECTOR(self);
    int i, size;
    gboolean retval = FALSE;

    size = v->size;
    for (i = 0; !retval && i < size; i++) {
	retval = nxp_element_depends_on(_element_vector_index(v, i), variable, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    break;
	}
    }
    return retval;
}

static NxpElement* substitute_func(NxpElement *self, NxpElement *var, NxpElement *el, NxpEvalContext *context, GError **error)
{
    NxpElement *tmp;
    NxpElementVector *v = NXP_ELEMENT_VECTOR(self), *retval;
    int i, size;
    GError *err = NULL;

    size = v->size;
    retval = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (i = 0; i < size; i++) {
	tmp = nxp_element_substitute(_element_vector_index(v, i), var, el, context, &err);
	if (err) {
	    NXP_ELEMENT_UNREF(NXP_ELEMENT(retval));
	    g_propagate_error(error, err);
	    return NULL;
	}
	_element_vector_index(retval, i) = tmp;
    }
    return NXP_ELEMENT(retval);
}


static NxpElement* map_op_simplify(NxpOp op,
			  NxpElement *a, NxpElement *b,
			  NxpEvalContext *context, GError **error) {
    NxpElementVector *result;
    gint i;
    GError *err = NULL;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(b), NULL);

    if (NXP_ELEMENT_VECTOR(a)->size != NXP_ELEMENT_VECTOR(b)->size) {
	// TODO -- GERROR SIZES DIFERENT
    } else {
	result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(NXP_ELEMENT_VECTOR(a)->size,
							   NULL));
	for (i = 0; i < NXP_ELEMENT_VECTOR(a)->size; i++)
	{
	    _element_vector_index(result, i) = nxp_element_tree_simplify_op(
			    op,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, &err);
	    if (err) {
		NXP_ELEMENT_UNREF(NXP_ELEMENT(result));
		g_propagate_error(error, err);
		return NULL;
	    }
	}
	return NXP_ELEMENT(result);
    }
    return NULL;
}

static NxpElement* map_op_eval(NxpOp op,
			       NxpElement *a, NxpElement *b,
			       NxpEvalContext *context, GError **error) {
    NxpElementVector *result;
    gint i;

    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(a), NULL);
    g_return_val_if_fail(NXP_ELEMENT_IS_VECTOR(b), NULL);

    if (NXP_ELEMENT_VECTOR(a)->size != NXP_ELEMENT_VECTOR(b)->size) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The given vectors are of different sizes"));
	return NULL;
    } else {
	result = NXP_ELEMENT_VECTOR(nxp_element_vector_new(NXP_ELEMENT_VECTOR(a)->size,
							   NULL));
	for (i = 0; i < NXP_ELEMENT_VECTOR(a)->size; i++)
	{
	    _element_vector_index(result, i) = nxp_element_tree_eval_op(
			    op,
			    _element_vector_index(NXP_ELEMENT_VECTOR(a), i),
			    _element_vector_index(NXP_ELEMENT_VECTOR(b), i),
			    context, error);
	}
	return NXP_ELEMENT(result);
    }
    return NULL;
}

static gboolean truth_func(NxpElement *v, NxpEvalContext *context, GError **error)
{
    int i;

    for (i = 0; i < NXP_ELEMENT_VECTOR(v)->size; i++)
	if (nxp_element_is_true(_element_vector_index(NXP_ELEMENT_VECTOR(v), i),
				context, error))
	    return TRUE;
    return FALSE;
}

int nxp_element_vector_get_type(void)
{
    static int nxp_element_type_vector = 0;

    if (nxp_element_type_vector == 0) {
	nxp_element_type_vector = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_vector,
					    sizeof(NxpElementTypeInfo),
					    0,
					    sizeof(NxpElementTypeInfo));
    }
    return nxp_element_type_vector;
}

void nxp_element_vector_type_init() 
{
    typeinfo = nxp_element_type_get_info(nxp_element_vector_get_type());
    typeinfo->name = g_strdup("vector");
    typeinfo->destroy_func = nxp_destroy_vector;
    typeinfo->eval_func = nxp_eval_vector;
    typeinfo->simplify_func = nxp_simplify_vector;
    typeinfo->derive_func = NULL;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = repr_func;
    typeinfo->truth_func = truth_func;
    typeinfo->iterator_func = iterator_func;
    typeinfo->dependson_func = dependson_func;
    typeinfo->substitute_func = substitute_func;

    BIND(NXP_OP_PLUS);
    BIND(NXP_OP_MINUS);
    BIND(NXP_OP_STAR);
    BIND(NXP_OP_SLASH);
    BIND_LEFT(minus, minus_simp, NXP_OP_MINUS_LEFT);
    //    NXP_OP_HAT

    /// not implementable ///////
    //
    //    NXP_OP_DOT3
    //    NXP_OP_PLUS2_LEFT
    //    NXP_OP_MINUS2_LEFT
    //    NXP_OP_PLUS2_RIGHT
    //    NXP_OP_MINUS2_RIGHT
    /////////////////////////////

    // >,<,>= and <= works with lexicographical order
    // Not complete, we should have {1, 2, 3} > {1, 2}
    BIND_FUNC( gt, gt, NXP_OP_GT);
    BIND_FUNC( le, le, NXP_OP_LT_EQ);
    BIND_FUNC( lt, lt, NXP_OP_LT);
    BIND_FUNC( ge, ge, NXP_OP_GT_EQ);

    // == and != compares ALL
    BIND_FUNC( eq, eq, NXP_OP_EQ2);
    BIND_FUNC( ne, NULL, NXP_OP_EXCL_EQ);

    // These two compares sizes, only
    BIND_FUNC( and, and, NXP_OP_SLASH_BSLASH);
    BIND_FUNC( or,  or,  NXP_OP_BSLASH_SLASH);
    BIND_LEFT( not, not, NXP_OP_TILDE_LEFT);

    ///// not implementable ///
    //
    //    NXP_OP_EXCL_RIGHT
    //    NXP_OP_EQ
    //    NXP_OP_COLON_EQ
    ///////////////////////////
}
