#include <gmp.h>
#include "nxp_kernel.h"
#include "ell_curve.h"

gboolean
_invert_array(mpz_t *xr, mpz_t *x, mpz_t n)
{
    mpz_t c[NXP_NB_PAR_PTS], inv, tmp;
    int k;

    mpz_init_set(c[0], x[0]);
    mpz_init(inv);
    for (k = 1; k < NXP_NB_PAR_PTS; k++) {
	mpz_init(c[k]);
	mpz_mul(c[k], c[k - 1], x[k]);
	mpz_mod(c[k], c[k], n);
    }
    if (mpz_invert(inv, c[NXP_NB_PAR_PTS - 1], n)) {
	mpz_init(tmp);
	for (k = NXP_NB_PAR_PTS - 1; k > 0; k--) {
	    mpz_mul(tmp, inv, c[k - 1]);
	    mpz_mul(inv, inv, x[k]);
	    mpz_mod(inv, inv, n);
	    mpz_mod(xr[k], tmp, n);
	}
	mpz_set(xr[0], inv);
	for (k = 0; k < NXP_NB_PAR_PTS; k++)
	    mpz_clear(c[k]);
	mpz_clear(inv);
	mpz_clear(tmp);

	return TRUE;
    }
    else {
	for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	    mpz_gcd(inv, x[k], n);
	    if (mpz_cmp_ui(inv, 1) > 0 && mpz_cmp(inv, n) < 0) {
		mpz_set(xr[0], inv);
		break;
	    }
	}
    }
    for (k = 0; k < NXP_NB_PAR_PTS; k++)
	mpz_clear(c[k]);
    mpz_clear(inv);

    return FALSE;
}

/**
 * ell_curve_add_point:
 * @x3, @y3: coordinates of the output point
 * @x1, @y1: coordinates of the first point
 * @x2, @y2: coordinates of the second point
 * @n: We compute this addition modulo @n which is an integer.
 *
 * This function computes the coordinates (x3,y3) of the point (x1,y1)+(x2,y2)
 * where those two points are on an elliptic curve in the ring Z/nZ. This
 * addition is not the trivial one at all ! The goal is that from (x1,y1) and
 * (x2,y2) we can have another point (x3,y3) on the curve too with coordinates in the ring Z/nZ. Sometimes this last point doesn't exist. In that case, we have a divisor of n :-), if we are very unlucky, the divisor found is n itself, but
 * this almost never occurs.
 *
 * Return value: If (x3,y3) exists, the function return TRUE, else it returns FALSE and x3 contains a divisor of n.
 **/
gboolean
ell_curve_add_points(mpz_t *x3, mpz_t *y3, mpz_t *x1, mpz_t *y1,
		     mpz_t *x2, mpz_t *y2, mpz_t n)
{
    mpz_t xr[NXP_NB_PAR_PTS], yr[NXP_NB_PAR_PTS];
    int k;

    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	mpz_init(xr[k]);
	mpz_init(yr[k]);
	mpz_sub(xr[k], x2[k], x1[k]);
    }
    if (_invert_array(yr, xr, n)) {
	mpz_t lambda;

	mpz_init(lambda);
	for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	    mpz_sub(lambda, y2[k], y1[k]);
	    mpz_mul(lambda, lambda, yr[k]);
	    mpz_mod(lambda, lambda, n); /* lambda is done */

	    mpz_mul(xr[k], lambda, lambda);
	    mpz_sub(xr[k], xr[k], x1[k]);
	    mpz_sub(xr[k], xr[k], x2[k]);
	    mpz_mod(xr[k], xr[k], n);		    /* xr is done */

	    mpz_sub(yr[k], x1[k], xr[k]);
	    mpz_mul(yr[k], yr[k], lambda);
	    mpz_sub(yr[k], yr[k], y1[k]);
	    mpz_mod(yr[k], yr[k], n);	    /* yr is done */
	    mpz_set(x3[k], xr[k]);
	    mpz_set(y3[k], yr[k]);
	    mpz_clear(xr[k]);
	    mpz_clear(yr[k]);
	}
	mpz_clear(lambda);
	return TRUE;
    }
    else {
	mpz_set(x3[0], yr[0]);
	for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	    mpz_clear(xr[k]);
	    mpz_clear(yr[k]);
	}
	return FALSE;
    }
}

    

/**
 * ell_curve_double_point:
 * @x3, @y3: coordinates of the output point
 * @x1, @y1: coordinates of the first point
 * @b: the equation of the elliptic curve is y^2=x^3+ax^2+bx+c. Here, we don't need c and a is null.
 * @n: We compute this addition modulo @n which is an integer.
 *
 * This function computes the coordinates (x3,y3) of the point (x1,y1)+(x2,y2)
 * where those two points are on an elliptic curve in the ring Z/nZ. This
 * addition is not the trivial one at all ! The goal is that from (x1,y1) and
 * (x2,y2) we can have another point (x3,y3) on the curve too with coordinates in the ring Z/nZ. Sometimes this last point doesn't exist. In that case, we have a divisor of n :-), if we are very unlucky, the divisor found is n itself, but
 * this almost never occurs.
 *
 * Return value: If (x3,y3) exists, the function return TRUE, else it returns FALSE and x3 contains a divisor of n.
 **/
gboolean
ell_curve_double_point(mpz_t *x3, mpz_t *y3, mpz_t *x1, mpz_t *y1, int *b, mpz_t n)
{
    mpz_t xr[NXP_NB_PAR_PTS], yr[NXP_NB_PAR_PTS];
    int k;

    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	mpz_init(xr[k]);
	mpz_init(yr[k]);
	mpz_mul_2exp(xr[k], y1[k], 1);
    }
    if (_invert_array(yr, xr, n)) {
	mpz_t lambda;

	mpz_init(lambda);
	for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	    mpz_set(lambda, x1[k]);
	    mpz_mul(lambda, lambda, x1[k]);
	    mpz_mul_ui(lambda, lambda, 3);

	    mpz_add_ui(lambda, lambda, b[k]);

	    mpz_mul(lambda, lambda, yr[k]);
	    mpz_mod(lambda, lambda, n); /* lambda is done */

	    mpz_mul(xr[k], lambda, lambda);

	    mpz_sub(xr[k], xr[k], x1[k]);
	    mpz_sub(xr[k], xr[k], x1[k]);
	    mpz_mod(xr[k], xr[k], n);		    /* xr is done */

	    mpz_sub(yr[k], x1[k], xr[k]);
	    mpz_mul(yr[k], yr[k], lambda);
	    mpz_sub(yr[k], yr[k], y1[k]);
	    mpz_mod(y3[k], yr[k], n);	    /* yr is done */
	    mpz_clear(yr[k]);
	    mpz_set(x3[k], xr[k]);
	    mpz_clear(xr[k]);
	}
	mpz_clear(lambda);
	return TRUE;
    }
    else {
	mpz_set(x3[0], yr[0]);
	for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	    mpz_clear(xr[k]);
	    mpz_clear(yr[k]);
	}
	return FALSE;
    }
}

gboolean
ell_curve_mul_point(mpz_t *x3, mpz_t *y3, nxp_int_t l, mpz_t *x1, mpz_t *y1,
		    int *b, mpz_t n)
{
    mpz_t x2p[NXP_NB_PAR_PTS], y2p[NXP_NB_PAR_PTS], xr[NXP_NB_PAR_PTS], yr[NXP_NB_PAR_PTS];
    int k;

    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	mpz_init_set(x2p[k], x1[k]);
	mpz_init_set(y2p[k], y1[k]);
    }

    while ((l & 1) == 0) {
	if (!ell_curve_double_point(x2p, y2p, x2p, y2p, b, n)) {
	    mpz_set(x3[0], x2p[0]);
	    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
		mpz_clear(x2p[k]);
		mpz_clear(y2p[k]);
	    }
	    return FALSE;
	}
	l >>= 1;
    }

    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	mpz_init_set(xr[k], x2p[k]);
	mpz_init_set(yr[k], y2p[k]);
    }
    l >>= 1;

    while (l) {
	if (!ell_curve_double_point(x2p, y2p, x2p, y2p, b, n)) {
	    mpz_set(x3[0], x2p[0]);
	    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
		mpz_clear(x2p[k]);
		mpz_clear(y2p[k]);
	        mpz_clear(xr[k]);
		mpz_clear(yr[k]);
	    }
	    return FALSE;
	}
	if (l & 1) {
	    if (!ell_curve_add_points(xr, yr, xr, yr, x2p, y2p, n)) {
		mpz_set(x3[0], xr[0]);
		for (k = 0; k < NXP_NB_PAR_PTS; k++) {
		    mpz_clear(x2p[k]);
		    mpz_clear(y2p[k]);
		    mpz_clear(xr[k]);
		    mpz_clear(yr[k]);
		}
		return FALSE;
	    }
	}
	l >>= 1;
    }
    for (k = 0; k < NXP_NB_PAR_PTS; k++) {
	mpz_set(x3[k], xr[k]);
	mpz_set(y3[k], yr[k]);
	mpz_clear(xr[k]);
	mpz_clear(yr[k]);
	mpz_clear(x2p[k]);
	mpz_clear(y2p[k]);
    }
    return TRUE;
}
